from .download import download_zip, download_file
from pathlib import Path
down_path = Path(__file__).parent
xz_url = "addons_file"+"/"+down_path.name+"/"
download_file(xz_url + "ANIMAX_generators.blend",down_path)
download_file(xz_url + "ANIMAX_node_groups.blend",down_path)
download_zip(xz_url + "Animation_Presets.zip",down_path)
download_zip(xz_url + "Images.zip",down_path)
download_zip(xz_url + "Textures.zip",down_path)

if "bpy" in locals():
    import importlib
    if "animax_fx_ui" in locals():
        importlib.reload(animax_fx_ui)
    if "animax_geo_nodes" in locals():
        importlib.reload(animax_geo_nodes)
    if "animax_items_ui" in locals():
        importlib.reload(animax_items_ui)
    if "animax_shell_splitting" in locals():
        importlib.reload(animax_shell_splitting)
    if "animax_slicing" in locals():
        importlib.reload(animax_slicing)
    if "animax_voronoi_fracture" in locals():
        importlib.reload(animax_voronoi_fracture)
else:
    from . import animax_fx_ui as fxui
    from . import animax_items_ui as iui
    from . import animax_geo_nodes as nds
    from . import animax_shell_splitting
    from . import animax_slicing
    from . import animax_voronoi_fracture

import bpy, bmesh, os, json
from random import uniform, choice, shuffle, seed
from mathutils import Vector, Euler, Matrix
from bpy.props import *
from bpy.types import UIList, Panel, Operator, Menu, AddonPreferences, Header
from mathutils.bvhtree import BVHTree
from bpy_extras.io_utils import ImportHelper
from bpy.app.handlers import load_post, persistent

# Add-on info
bl_info = {
    "name": "ANIMAX",
    "author": "Monaime Zaim",
    "version": (2, 3, 0),
    "blender": (4, 1, 0),
    "location": "View3D > Tools > ANIMAX",
    "description": "Procedural animation tool.",
    "wiki_url": "https://blendermarket.com/products/animax---procedural-animation-system-/docs",
    "tracker_url": "https://blendermarket.com/products/animax---procedural-animation-system-",      
    "category": "Animation",
    "warning": ""
    }
    
# Margine between the nodes
x_margine, y_margine = [400.0, 500.0]
    
# Presets folder
addon_folder = os.path.dirname(__file__)
ng_dir = os.path.join(addon_folder, "ANIMAX_node_groups.blend")
generators_dir = os.path.join(addon_folder, "ANIMAX_generators.blend")
anim_presets = os.path.join(addon_folder, "Animation_Presets")
icons_dir = os.path.join(addon_folder, "Images")
clipboard_path = os.path.join(addon_folder, "Clipboard")
textures_path = os.path.join(addon_folder, "Textures")

# Previews Collection for the icons
preview_collections = {}

# All the Items
# type[ui_name, icon_name, ui_panel]
items_dict = {
    'COLLECTION': ['Collection', 'OUTLINER_COLLECTION', iui.coll_item_ui],
    'TEXT': ['Text', 'OUTLINER_OB_FONT', iui.text_item_ui],
    'NODE_GROUP': ['Node Group', 'NODETREE', iui.custom_ng_item_ui],
}
# All the Effects
# type[ui_name, function_node, ui_panel]
effects_dict = {
    'TRANSFORM': ['Transform', nds.transform_fx_node, fxui.transform_fx_ui],
    'NOISE': ['Noise', nds.noise_fx_node, fxui.noise_fx_ui],
    'FOLLOW_CURVE': ['Follow Curve', nds.follow_curve_fx_node, fxui.follow_curve_fx_ui],
    'ATTRACT': ['Attract', nds.attract_fx_node, fxui.attract_fx_ui],
    'EXPLODE': ['Explode', nds.explode_fx_node, fxui.explode_fx_ui],
    'REPULSE': ['Repulse', nds.repulse_fx_node, fxui.repulse_fx_ui],
    'PROJECT': ['Project', nds.project_fx_node, fxui.project_fx_ui],
    'COLLAPSE': ['Collapse', nds.collapse_fx_node, fxui.collapse_fx_ui],
    'ORBIT': ['Orbit', nds.orbit_fx_node, fxui.orbit_fx_ui],
    'WAVE': ['Wave', nds.wave_fx_node, fxui.wave_fx_ui],
    'ROBOTIC': ['Robotic', nds.robotic_fx_node, fxui.robotic_fx_ui],
    'EMPTY': ['Empty', nds.empty_fx_node, fxui.empty_fx_ui],
}

# Offset and Timing properties (common for all the effects)
offset_props = [
    'offset',
    'offset_direction',
    'invert_direction',
    'offset_falloff',
    'random_offset',
    'random_offset_seed',
    'effector_type',
    'effector_object',
    'use_effector_scale',
    'effector_distance',
    'effector_falloff',
    'texture_type',
    'texture_location',
    'texture_rotation',
    'texture_scale',
    'noise_texture_scale',
    'noise_texture_detail',
    'noise_texture_roughness',
    'noise_texture_distortion',
    'noise_texture_frequency',
    'image_texture',
    'control_time',
    'time',
    'start_frame',
    'duration',
    'invert_time',
    'repeat_time',
    'ping_pong',
    'easing_type',
    'easing',
    'easing_smooth_step',
    'easing_frequency',
    'easing_amplitude',
]

# The properties corresponding to each effect
fx_props = {
    'TRANSFORM':[
        'replace_location',
        'replace_rotation',
        'replace_scale',
        'explode_location',
        'location',
        'random_location',
        'random_location_seed',
        'rotation',
        'random_rotation',
        'random_rotation_seed',
        'scale',
        'random_scale',
        'random_scale_seed',
        'scale_fade_out',
        'fade_out_smooth_step',
        'uniform_scale',
        'uniform_scale_value',
        'uniform_random_scale'
    ],
    'NOISE':[
        'location',
        'rotation',
        'align_rotation',
        'axis_enum',
        'scale',
        'uniform_scale',
        'uniform_scale_value',
        'scale_fade_out',
        'fade_out_smooth_step',
        'continuous_animation',
        'noise_frequency',
        'noise_scale',
        'noise_detail',
        'noise_roughness',
        'noise_distortion',
        'random_seed'
    ],
    'FOLLOW_CURVE':[
        'scale_fade_out',
        'fade_out_smooth_step',
        'follow_object',
        'follow_resolution',
        'follow_resolution',
        'follow_reverse_curve',
        'follow_world_position',
        'align_rotation',
        'axis_enum',
        'invert_axis',
        'num_turns'
    ],
    'ATTRACT':[
        'scale_fade_out',
        'fade_out_smooth_step',
        'attractor',
        'collapse_axis',
        'align_rotation',
        'axis_enum',
        'invert_axis'
    ],
    'EXPLODE':[
        'scale_fade_out',
        'fade_out_smooth_step',
        'velocity',
        'gravity',
        'random_location',
        'random_location_seed',
        'time_multiplier'
    ],
    'REPULSE':[
        'scale_fade_out',
        'fade_out_smooth_step',
        'repulse_method',
        'repulsor',
        'force',
        'distance'
    ],
    'PROJECT':[
        'scale_fade_out',
        'fade_out_smooth_step',
        'factor',
        'col_ob_switch',
        'target_ob',
        'target_col',
        'align_rotation',
        'axis_enum',
        'ray_direction',
        'ray_length',
        'fade_out_missed',
    ],
    'COLLAPSE':[
        'scale_fade_out',
        'fade_out_smooth_step',
        'collapse_axis',
        'collapse_offset',
        'collapse_rotation'
    ],
    'ORBIT':[
        'scale_fade_out',
        'fade_out_smooth_step',
        'center_ob',
        'rotation'
    ],
    'WAVE':[
        'location',
        'rotation',
        'scale',
        'scale_fade_out',
        'fade_out_smooth_step',
        'continuous_animation',
        'coord_rotation',
        'noise_frequency',
        'noise_scale',
        'noise_detail',
        'details_scale',
        'noise_roughness',
        'noise_distortion',
        'wave_phase'
    ],
    'ROBOTIC':[
        'rotation',
        'distance_vec',
        'scale_fade_out',
        'fade_out_smooth_step',
        'random_seed'
    ],
    'EMPTY':[
        'scale_fade_out',
        'fade_out_smooth_step'
    ],
}
# Properties of the exposed props
exposed_fx_prop = [
    'fx',
    'id',
    'original_name',
    'ui_name',
    'description',
]
# Properties of the items
# Used only for duplicating the item
items_props = [
    'collection',
    'control_time',
    'custom_ng',
    'frame',
    'fx_enabled',
    'merge_by_distance',
    'realize_instances',
    'delete_invisible',
    'use_local_location',
    'use_local_rotation',
    'use_local_scale'
    ]
# update the active preset displayed when some properties changes
def update_presets(self, context):
    presets_enum = get_all_presets(context)
    context.scene.animax_fx_presets = presets_enum[0][0]
    return None

def update_drivers():
    for node_group in bpy.data.node_groups:
        if node_group.type != 'GEOMETRY':
            continue
        if not hasattr(node_group, 'animation_data'):
            continue
        animation_data = node_group.animation_data
        if not hasattr(animation_data, 'drivers'):
            continue
        for d in animation_data.drivers:
            driver = d.driver
            if driver.type == 'SCRIPTED':
                driver.expression = driver.expression

# On blend load Handler
@persistent
def animax_load_handler(dummy):
    # Load the properties if necessary
    scn = bpy.context.scene
    if 'animax_items' in scn and not hasattr(scn, 'animax_items'):
        if not 'coll' in scn['animax_items']:
            return
        if not len(scn['animax_items']['coll']):
            return
        load_anim_props()
        update_drivers()
        print('ANIMAX Properties Loaded')
################################################################################
############################## Preferences #####################################
################################################################################

class ANIMAX_Preferences(AddonPreferences):
    bl_idname = __package__
    
    preferences_tabs : EnumProperty(
        items = (
            ('PRESETS', 'Presets', 'Your custom presets.', 'FILE_FOLDER', 0),
            ('DOC', 'Documentation', 'Documentation of the add-on.', 'DOCUMENTS', 1),
            ('MORE', 'More from us', 'More of our tools.', 'URL', 2),
        ),
        name = 'Preferences Tabs',
        default = 'PRESETS',
        description = 'Preferences Tabs'
    )
    presets_path : StringProperty(
        name="Custom Presets Folder",
        update = update_presets,
        description = 'The folder where your custom presets are stored.\n'
            "Don't forget to save the preferences after setting it up",
        subtype='DIR_PATH'
    )
    remove_button : BoolProperty(
        name = 'Romove Buttons',
        default = True,
        description = 'Display the remove parameters buttons'
    )
    move_button : BoolProperty(
        name = 'Move Buttons',
        default = True,
        description = 'Display the move parameters buttons'
    )
    info_button : BoolProperty(
        name = 'Info Buttons',
        default = True,
        description = 'Display the Info buttons'
    )
    
    def draw(self, context):
        scn = context.scene
        layout = self.layout
        col = layout.column()
        row = col.row()
        row.prop(self, 'preferences_tabs', expand = True)
        ################### Presets ###################
        if self.preferences_tabs == 'PRESETS':
            box = layout.box()
            col = box.column()
            col.prop(self, 'presets_path')
            if not self.presets_path.strip():
                col.label(text = "Please select a folder", icon = 'INFO')
            elif not os.path.exists(self.presets_path):
                col.label(text = "Wrong Folder Path", icon = 'ERROR')
            else:
                len_slicing, len_animation = num_custom_presets(context)
                box = col.box()
                box.label(text = "Available Presets in this folder", icon = 'INFO')
                box.label(text = "Animation Preset: " + str(len_animation))
                box.label(text = "Slicing Preset: " + str(len_slicing))
        ################### Documentation ###################
        elif self.preferences_tabs == 'DOC':
            box = layout.box()
            box.operator("animax.open_manual", icon = 'TEXT')
            ico = get_custom_icon_id('yt_logo.png')
            box.operator("animax.video_tuts", icon_value = ico)
            
        ################### More from us ###################
        elif self.preferences_tabs == 'MORE':
            box = layout.box()
            ico = get_custom_icon_id('coa.png')
            box.label(text = 'Check out our stores', icon_value=ico)
            
            ico = get_custom_icon_id('gumroad_logo.png')
            url = 'https://codeofart.gumroad.com/'
            text = 'Our store in Gumroad'
            box.operator("wm.url_open", text = text, icon_value=ico).url = url
            
            ico = get_custom_icon_id('blender_market_logo.png')
            url = 'https://blendermarket.com/creators/monaime'
            text = 'Our store in the Blender Market'
            box.operator("wm.url_open", text = text, icon_value=ico).url = url
            
# Add-on Preferences
def prefs(context):
    # addon = context.preferences.addons.get(__name__)
    # if addon is not None:
    # return addon.preferences
    # return None
    return context.preferences.addons[__package__].preferences

# Custom presets folder
def custom_presets_dir(context):
    preferences = prefs(context)
    if preferences is None:
        return None
    if not os.path.exists(preferences.presets_path):
        return None
    return preferences.presets_path

############################################################################################
######################################## Utils #############################################
############################################################################################

# get icon ID
def get_custom_icon_id(image_name):
    pcoll = preview_collections["icons"]
    icon = pcoll.get(image_name)
    if icon is not None:
        return icon.icon_id
    print('Error with the image:', image_name)
    return 0
    

# Create a collection for the geo node containers
def containers_coll(context):
    items = context.scene.animax_items
    coll = items.containers_coll
    if coll is None:
        coll_name = 'ANIMAX_Containers'
        main_coll = context.scene.collection
        collections = bpy.data.collections
        scn_colls = [c.name for c in main_coll.children_recursive]
        
        if coll_name in scn_colls:
            coll = collections[coll_name]
        else:
            coll = collections.new(coll_name)
            main_coll.children.link(coll)
        items.containers_coll = coll
    return coll

# Get a new unique name for a property
def get_prop_name(name, names):
    newname = name    
    i = 2
    while newname in names:
        newname = name + str(i)
        i += 1
    return newname

# Get the index of a property in a group
def get_item_idx(items, name):
    for idx, item in enumerate(items):
        if item.name == name:
            return idx
    return None

# Add the object container for the geo node modifier
def add_geo_node_container(context):
    coll = containers_coll(context)
    me = bpy.data.meshes.new("ANIMAX_Container_Mesh")
    name = 'ANIMAX_GeoNode_Container'
    names = [o.name for o in bpy.data.objects]
    name = get_prop_name(name, names)
    ob = bpy.data.objects.new(name, me)
    mod = ob.modifiers.new('ANIMAX', 'NODES')
    ng = bpy.data.node_groups.new('ANIMAX_Geo_NG', 'GeometryNodeTree')
    mod.node_group = ng
    nodes = ng.nodes
    if len(nodes):
        nodes.clear()
    coll.objects.link(ob)
    return ob

# Items collection
def items_coll(context):
    coll = context.scene.animax_items.coll
    index = context.scene.animax_items.index
    return coll, index

# Add Item
def add_item(context, item_type):
    scn = context.scene
    coll = scn.animax_items.coll
    names = coll.keys()
    ui_names = [i.ui_name for i in coll]
    item = coll.add()
    item.name = get_prop_name('Item', names)
    item.ui_name = get_prop_name(items_dict[item_type][0], ui_names)
    item.type = item_type
    container = add_geo_node_container(context)
    item.geo_node_container = container
    scn.animax_items.index = len(scn.animax_items.coll) -1
    coll.update()
    generate_items_node_tree(context, item)
    return item
    
# Remove the selected Item
def remove_item(context):
    scn = context.scene
    coll, index = items_coll(context)
    item = coll[index]
    node_group = item_ng(item)
    container = item.get('geo_node_container')
    collection = scn.animax_items.get('containers_coll')
    
    for fx in item.fx:
        delete_fx_node(item, fx)
        
    coll.remove(index)
    
    if index >0:
        scn.animax_items.index = (index - 1)
    if not container:
        return
    try:
        mesh = container.data
        bpy.data.meshes.remove(mesh)
        if node_group is not None:
            node_groups = bpy.data.node_groups
            if not node_group.users:
                node_groups.remove(node_group)
        if not len(scn.animax_items.coll):
            if collection is not None and not len(collection.all_objects):
                bpy.data.collections.remove(collection)
    except Exception as e:
        print('Error while deleting an item:', e)
    
# Load and register the animation properties
def load_anim_props():
    file = 'animax_animation_properties.py'
    texts = bpy.data.texts
    text = texts.get(file)
    if text:
        texts.remove(text)
    text = texts.load(os.path.join(addon_folder, file))
    text.use_module = True
    exec(text.as_string(), {})

# delete the effect node
def delete_fx_node(item, fx):
    container = item.get('geo_node_container')
    if container is None:
        return None
    mod = container.modifiers.get('ANIMAX')
    if mod is None:
        return None
    nodes = mod.node_group.nodes
    fx_node = nodes.get(fx.name)
    if fx_node is None:
        print('Effect Node Not Found!')
        return None
    
    if hasattr(fx_node, 'node_tree'):
        node_group = fx_node.node_tree
    nodes.remove(fx_node)
    
    try:
        bpy.data.node_groups.remove(node_group)
    except Exception as e:
        print(str(e))
        
# Clear node links
def clear_node_links(node, links):
    #inputs
    for i in node.inputs:
        if not hasattr(i, 'links'):
            continue
        if not len(i.links):
            continue
        links.remove(i.links[0])
    #outputs
    for i in node.outputs:
        if not hasattr(i, 'links'):
            continue
        if not len(i.links):
            continue
        for j in i.links:
            links.remove(j)
            
# Find the node-group of the active item
def item_ng(item):
    container = item.get('geo_node_container')
    if container is None:
        return None
    mod = container.modifiers.get('ANIMAX')
    if mod is None:
        return None
    if hasattr(mod, 'node_group'):
        return mod.node_group
    return None

# Find the effect node
def get_fx_node(item, fx):
    ng = item_ng(item)
    if ng is None:
        return None
    fx_node = ng.nodes.get(fx.name)
    return fx_node
    
# Find the curve node for an effect
def get_curve_node(item, fx, type = 'EASING'):
    ng = item_ng(item)
    if ng is None:
        return None
    curve = None
    fx_node = ng.nodes.get(fx.name)
    if fx_node is None:
        return None
    if type == 'EASING':
        curve = fx_node.node_tree.nodes.get('Easing_Curve')
    elif type == 'ROTATION':
        curve = fx_node.node_tree.nodes.get('Rotation_Curve')
    return curve

# store the curve data
def get_curve_node_data(curve_node):
    if not curve_node.type == 'CURVE_FLOAT':
        print(curve_node.name + ' Not a Float Curve Node')
        return []
    #curve_data = ['use_clip':Bool, 'points':[point_location, handle_type]]
    curve_data = [curve_node.mapping.use_clip, []]
    points = curve_node.mapping.curves[0].points
    for p in points:
        point_data = [p.location[:], p.handle_type]
        curve_data[1].append(point_data)
    return curve_data
        
# apply the curve data
def set_curve_node_data(curve_node, curve_data):
    if not curve_node.type == 'CURVE_FLOAT':
        print(curve_node.name + ' Not a Float Curve Node')
        return
    if not curve_data:
        print('No Curve Data')
        return
    points = curve_node.mapping.curves[0].points
    # Add the missing points
    dif = len(curve_data[1]) - len(points)
    for p in range(dif):
        points.new(1.0, 1.0)
    # set the location and handle-type
    for i, p in enumerate(curve_data[1]):
        points[i].location = p[0]
        points[i].handle_type = p[1]
    curve_node.mapping.use_clip = curve_data[0]
        
# Check if the effect is additive with the location
def is_fx_loc_additive(fx):
    if fx.type == 'PROJECT':
        return False
    elif fx.type == 'ATTRACT':
        if fx.factor > 0:
            return False
    elif fx.type == 'TRANSFORM':
        if not fx.explode_location and fx.replace_location:
            return False
    elif fx.type == 'FOLLOW_CURVE':
        if fx.follow_world_position:
            return False
    return True

# Check if the effect is additive with the rotation
def is_fx_rot_additive(fx):
    if fx.type in ['ATTRACT', 'FOLLOW_CURVE', 'PROJECT', 'NOISE']:
        if fx.align_rotation:
            return False
    elif fx.type == 'TRANSFORM':
        if fx.replace_rotation:
            return False
    return True

# Check if the effect is additive with the scale
def is_fx_scale_additive(fx):
    if fx.scale_fade_out:
        return False
    elif fx.type == 'TRANSFORM':
        if fx.replace_scale:
            return False
    elif fx.type == 'PROJECT':
        if fx.fade_out_missed:
            return False
    return True

# Check if the container is valid
def is_container_valid(item):
    container = item.get('geo_node_container')
    if container == None:
        return False
    if not container.users_collection:
        return False
    return True

def check_mesh(mesh):
    elements = {
        'vertices': [],
        'edges': [],
        'polygons': [],
    }
    info = dict()
    bm = bmesh.new()
    bm.from_mesh(mesh)
    # detect intersecting faces
    bm_BVHtree = BVHTree.FromBMesh(bm, epsilon=0.00001)
    inter = bm_BVHtree.overlap(bm_BVHtree)
    intersecting_faces = {f[0] for f in inter}
    elements['polygons'].extend(intersecting_faces)
    info['Intersecting Faces'] = len(intersecting_faces)
    # detect doubles
    doubles = bmesh.ops.find_doubles(bm, verts=bm.verts, dist=0.0001)
    double_verts = []
    for k, v in doubles['targetmap'].items():
        double_verts.extend([k.index, v.index])
    elements['vertices'].extend(double_verts)
    info['Overlapping Vertices'] = len(double_verts)
    # detect non manifold edges
    non_manifold_edges = [e.index for e in bm.edges if len(e.link_faces) != 2]
    elements['edges'].extend(non_manifold_edges)
    info['Non Manifold Edges'] = len(non_manifold_edges)
    # detect loose vertices
    loose_verts = [v.index for v in bm.verts if not v.link_edges]
    elements['vertices'].extend(loose_verts)
    info['Loose Vertices'] = len(loose_verts)
    bm.free()
    return info, elements

####################################### Drivers Utils ######################################

# Add property variable for the driver
def add_prop_var(driver, name, id_type, id, path):
    var = driver.driver.variables.new()
    var.name = name
    var.type = 'SINGLE_PROP'
    var.targets[0].id_type = id_type
    var.targets[0].id = id
    var.targets[0].data_path = path
    
# Convert the data-path to string
def prop_to_string(item, fx, prop):
    s = 'animax_items.coll["'+item+'"].fx["'+fx+'"].'+prop
    return s
    
############################################################################################
######################################## Nodes #############################################
############################################################################################
# Effects node group
def effects_ng():
    node_groups = bpy.data.node_groups
    ng = node_groups.new('ANIMAX_Effects_NG', 'GeometryNodeTree')
    
    nodes = ng.nodes
    links = ng.links
    loc = [0.0, 0.0]
    x_margine, y_margine = (400.0, 300.0)
        
    #Adding new nodes
    #Inputs
    group_input = nodes.new('NodeGroupInput')
    group_input.name = 'Group Input'    
    group_input.location = loc
    ng.interface.new_socket(socket_type='NodeSocketFloat', name='Time', in_out='INPUT')
    ng.interface.new_socket(socket_type='NodeSocketVector', name='Previous Location', in_out='INPUT')
    ng.interface.new_socket(socket_type='NodeSocketVector', name='Previous Rotation', in_out='INPUT')
    prev_scale = ng.interface.new_socket(socket_type='NodeSocketVector', name='Previous Scale', in_out='INPUT')
    ng.interface.new_socket(socket_type='NodeSocketGeometry', name='Instances', in_out='INPUT')
    ng.interface.new_socket(socket_type='NodeSocketFloat', name='Frame', in_out='INPUT')
    ng.interface.new_socket(socket_type='NodeSocketVector', name='Position', in_out='INPUT')
    ng.interface.new_socket(socket_type='NodeSocketVector', name='Center Vectors', in_out='INPUT')
    ng.interface.new_socket(socket_type='NodeSocketFloat', name='Direction', in_out='INPUT')

    prev_scale.default_value = (1.0, 1.0, 1.0)
    loc[0] += x_margine*6
    # Outputs
    group_output = nodes.new('NodeGroupOutput')
    group_output.name = 'Output'
    ng.interface.new_socket(socket_type='NodeSocketFloat', name='Time', in_out='OUTPUT')
    ng.interface.new_socket(socket_type='NodeSocketVector', name='Location', in_out='OUTPUT')
    ng.interface.new_socket(socket_type='NodeSocketVector', name='Rotation', in_out='OUTPUT')
    ng.interface.new_socket(socket_type='NodeSocketVector', name='Scale', in_out='OUTPUT')
    
    group_output.location = loc
    return ng

# Effect Node
def effect_node(context, nodes, item, fx, fx_index):
    node_groups = bpy.data.node_groups
    fx_node = nodes.get(fx.name)
    if fx_node is None:
        fx_node = nodes.new(type="GeometryNodeGroup")
        ng = effects_ng()
        fx_node.node_tree = ng
        fx_node.name = fx.name
        # driver for the frame
        driver = fx_node.inputs[5].driver_add('default_value')
        path = 'animax_items.coll["'+item.name+'"].frame'
        add_prop_var(driver, 'prop', 'SCENE', context.scene, path)
        driver.driver.expression = 'prop'
        
    ng_nodes = fx_node.node_tree.nodes
    ng_links = fx_node.node_tree.links
    
    fx_ng = effects_dict[fx.type][1](context, ng_nodes, ng_links, item, fx, ng_dir)
    fx_ng.location = (x_margine*3, 0.0)
    
    ng_input = ng_nodes['Group Input']
    ng_output = ng_nodes['Output']
    
    # Assign Extra properties on update
    if fx.type == 'REPULSE':
        fx_ng.inputs[8].default_value = fx.repulsor
    elif fx.type == 'ORBIT':
        fx_ng.inputs[5].default_value = fx.center_ob
    elif fx.type == 'PROJECT':
        fx_ng.inputs[7].default_value = fx.target_ob
        fx_ng.inputs[8].default_value = fx.target_col
    elif fx.type == 'FOLLOW_CURVE':
        fx_ng.inputs[5].default_value = fx.follow_object
    elif fx.type == 'ATTRACT':
        fx_ng.inputs[8].default_value = fx.attractor
    
    # Connect Align Rotation node to the output
    if fx.type in ['ATTRACT', 'FOLLOW_CURVE', 'PROJECT', 'NOISE']:
        if fx.align_rotation:
            align_euler = ng_nodes.get('Align_Euler')
            ng_links.new(align_euler.outputs[0], ng_output.inputs[2])
        else:
            ng_links.new(fx_ng.outputs['Rotation'], ng_output.inputs[2])
        
    # offset node
    offset = None
    if fx.sync and fx_index > 0:
        time_socket = ng_input.outputs['Time']
    elif fx.offset == 'POSITION':
        offset = nds.position_node(context, item, ng_nodes, fx, ng_dir)
        ng_links.new(ng_input.outputs['Instances'], offset.inputs['Instances'])
        ng_links.new(ng_input.outputs['Direction'], offset.inputs['Direction'])
        ng_links.new(ng_input.outputs['Frame'], offset.inputs['Frame'])
        ng_links.new(ng_input.outputs['Position'], offset.inputs['Position'])
        time_socket = offset.outputs[0]
    elif fx.offset == 'EFFECTOR':
        offset = nds.effector_node(context, item, ng_nodes, fx, ng_dir)
        ng_links.new(ng_input.outputs['Position'], offset.inputs['Position'])
        ng_links.new(ng_input.outputs['Previous Location'], offset.inputs['Previous Location'])
        offset.inputs[4].default_value= fx.effector_object
        time_socket = offset.outputs[fx.effector_type]
    elif fx.offset == 'TEXTURE':
        offset = nds.texture_node(context, item, ng_nodes, fx, ng_dir)
        ng_links.new(ng_input.outputs['Instances'], offset.inputs['Instances'])
        ng_links.new(ng_input.outputs['Position'], offset.inputs['Position'])
        ng_links.new(ng_input.outputs['Frame'], offset.inputs['Frame'])
        time_socket = offset.outputs[0]
        if fx.texture_type == 'IMAGE':
            if offset.inputs[13].default_value != fx.image_texture:
                offset.inputs[13].default_value = fx.image_texture
            
    if offset is not None:
        offset.location = (x_margine, 0.0)
    ng_links.new(time_socket, ng_output.inputs[0])
    if fx.type in ['FOLLOW_CURVE', 'ATTRACT', 'PROJECT', 'NOISE']:
        rot_curve = ng_nodes.get('Rotation_Curve')
        ng_links.new(time_socket, rot_curve.inputs[1])
    
    # easing Node
    easing = nds.easing_node(context, item, ng_nodes, fx, ng_dir)
    if easing is not None:
        if fx.easing_type == 'CURVE':
            ng_links.new(time_socket, easing.inputs[1])
        else:
            ng_links.new(time_socket, easing.inputs[0])
        easing.location = (x_margine*2, 0.0)
        time_socket = easing.outputs[0]
    ng_links.new(time_socket, fx_ng.inputs[0])
    
    return fx_node

# Get material node
def get_material_node(item):
    ng = item_ng(item)
    if ng is not None:
        return ng.nodes.get('ANIMAX_Set_Material_NG')

# Get text node
def get_text_node(item):
    ng = item_ng(item)
    if ng is not None:
        return ng.nodes.get('ANIMAX_Text')

# Get extrude node
def get_extrude_node(item):
    ng = item_ng(item)
    if ng is not None:
        return ng.nodes.get('ANIMAX_Extrude_NG')

############################################################################################
####################################### Effects ############################################
############################################################################################

# Generate the node tree for an item
def generate_items_node_tree(context, item):
    scn = context.scene
    effects = item.fx
    node_groups = bpy.data.node_groups
    container = item.geo_node_container
    mod = container.modifiers.get('ANIMAX')
    node_group = mod.node_group
    nodes = node_group.nodes
    links = node_group.links
    loc = [0.0, 0.0]
    pivot_socket = None
    
    # Scene Time Node
    scene_time = nds.scene_time_node(nodes)
    frame_socket = scene_time.outputs[1] if not item.control_time else None
    scene_time.location = loc
    
    loc[0] += x_margine
    # Geometry/Instances source
    if item.type == 'COLLECTION':
        # Collection Info Node
        geo = nds.coll_info_node(nodes)
        cont_coll = context.scene.animax_items.containers_coll
        if item.collection != cont_coll and geo.inputs[0].default_value != item.collection:
            geo.inputs[0].default_value = item.collection
    elif item.type == 'TEXT':
        # Text input node
        text = nds.text_input_node(nodes)
        text.location = [-x_margine, -y_margine]
        pivot_socket = text.outputs[3]
        # Fill curves node
        geo = nds.fill_curve_node(nodes)
        links.new(text.outputs[0], geo.inputs[0])
    elif item.type == 'NODE_GROUP':
        geo = nds.custom_ng_node(nodes, ng_dir, item)
        
    geo_socket = geo.outputs[0]
    geo.location = [0.0, -y_margine]
    
    # Collection Input Node
    coll_input = nds.coll_input_node(nodes, ng_dir)
    coll_input.location = loc
    links.new(geo_socket, coll_input.inputs[0])
    geo_socket = coll_input.outputs[0]
    
    loc[0] += x_margine
    
    # Apply Transform Node
    apply_trans = nds.apply_trans_node(nodes, ng_dir, item.name, scn)
    links.new(geo_socket, apply_trans.inputs['Instances'])
    if pivot_socket is not None:
        links.new(pivot_socket, apply_trans.inputs[4])
    
    # Realize Instances Node
    realize = nds.realize_node(nodes, item.name, ng_dir, scn)
    links.new(apply_trans.outputs[0], realize.inputs[0])
    geo = realize.outputs[0]
    
    # Text Items Add two more nodes
    if item.type == 'TEXT':
        # Extrude Node
        extrude = nds.extrude_node(nodes, item.name, ng_dir, scn)
        links.new(geo, extrude.inputs[0])
        geo = extrude.outputs[0]
    
    # Set material Node
    set_material = nds.set_material_node(nodes, ng_dir)
    links.new(geo, set_material.inputs[0])
    geo = set_material.outputs[0]
    
    # Group Output Node
    group_output = nds.group_output_node(nodes)
    if len(group_output.inputs) < 2:
         node_group.interface.new_socket(socket_type='NodeSocketGeometry', name='Geometry', in_out='OUTPUT')
        
    links.new(geo, group_output.inputs[0])
    
    # Clear the "FX" outputs
    items_tree = node_group.interface.items_tree
    sockets = [i for i in items_tree if i.item_type == 'SOCKET' and i.in_out == 'OUTPUT' and 'FX' in i.name]
    for socket in sockets:
        node_group.interface.remove(socket)
    
    # The Effects
    trans = ['Location', 'Rotation', 'Scale']
    prev_node = None
    time_sockets = {}
    for i, fx  in enumerate(effects):
        # Effect Node
        fx_node = effect_node(context, nodes, item, fx, i)
        clear_node_links(fx_node, links)
        fx_node.location = loc
        # fx : [socket, index]
        time_sockets[fx.name] = [fx_node.outputs[0], i]
        time_socket = None
        
        loc[0] += x_margine
        
        # Sync
        if fx.sync and fx.sync_with and len(time_sockets) and i != 0:
            sync_node = nodes.get(fx.sync_with)
            sync_id = time_sockets.get(fx.sync_with)
            if sync_node and sync_id:
                sync_index = sync_id[1]
                if i > sync_index:
                    time_socket = sync_id[0]
        
        # Links
        if len(fx_node.inputs[0].links):
            links.remove(fx_node.inputs[0].links[0])
        
        if time_socket:
            links.new(time_socket, fx_node.inputs[0])
        
        if frame_socket is not None:
            links.new(frame_socket, fx_node.inputs['Frame'])
        links.new(geo_socket, fx_node.inputs['Instances'])
        links.new(coll_input.outputs['Position'], fx_node.inputs['Position'])
        links.new(coll_input.outputs['Center Vectors'], fx_node.inputs['Center Vectors'])
        links.new(coll_input.outputs[fx.offset_direction], fx_node.inputs['Direction'])
            
        if fx.enabled:
            if prev_node:
                for l in trans:
                    links.new(prev_node.outputs[l], fx_node.inputs['Previous '+l])
            prev_node = fx_node
            
        # Store FX Attribute
        out = node_group.interface.new_socket(socket_type='NodeSocketFloat', name=fx.name, in_out='OUTPUT')
        if item.realize_instances:
            out.attribute_domain = 'POINT'
        else:
            out.attribute_domain = 'INSTANCE'
        id = out.identifier
        mod[id+'_attribute_name'] = fx.name
        fx_time_output = group_output.inputs.get(fx.name)
        if fx_time_output is not None:
            links.new(fx_node.outputs[0], fx_time_output)
    
    if prev_node is not None:
        for l in trans:
            links.new(prev_node.outputs[l], apply_trans.inputs[l])
    
    apply_trans.location = loc
    loc[0] += x_margine
    
    realize.location = loc
    loc[0] += x_margine
    
    if item.type == 'TEXT':
        extrude.location = loc
        loc[0] += x_margine
    
    set_material.location = loc
    loc[0] += x_margine
    
    group_output.location = loc    
        
        
# Add Effect
def add_fx(context, item, fx_type, ui_name):
    ui_names = [i.ui_name for i in item.fx]
    names = item.fx.keys()
    fx = item.fx.add()
    fx.name = get_prop_name('FX', names)
    fx.ui_name = get_prop_name(ui_name, ui_names)
    fx.type = fx_type
    item.fx_index = len(item.fx) -1
    generate_items_node_tree(context, item)

# Delete Selected Effect
def remove_fx(context, item):
    coll = item.fx
    index = item.fx_index
    fx = coll[index]
    delete_fx_node(item, fx)
    # Remove the exposed properties
    remove_exposed_fx_prop(item, fx)
    # remove the fx
    coll.remove(index)
    if index >0:
        item.fx_index = (index - 1)
    generate_items_node_tree(context, item)
    
# Get the data of an effect
def get_fx_data(item, fx):
    fx_data = {}
    fx_data['name'] = fx.name
    fx_data['ui_name'] = fx.ui_name
    fx_data['type'] = fx.type
    fx_data['enabled'] = fx.enabled
    fx_data['sync'] = fx.sync
    fx_data['sync_with'] = fx.sync_with
    # time offset data
    for p in offset_props:
        if not hasattr(fx, p):
            print(fx.type, ', Attribute Not Found:', p)
            continue
        fx_data[p] = getattr(fx, p)
    # get offset curve data
    curve_node = get_curve_node(item, fx, type = 'EASING')
    if curve_node is not None:
        curve_data = get_curve_node_data(curve_node)
        if curve_data:
            fx_data['OFFSET_CURVE'] = curve_data
    # effect data
    for p in fx_props[fx.type]:
        if not hasattr(fx, p):
            print('Property Not Found:', p)
            continue
        fx_data[p] = getattr(fx, p)
    # get rotation curve data
    if fx.type in ['FOLLOW_CURVE', 'ATTRACT', 'NOISE', 'PROJECT']:
        curve_node = get_curve_node(item, fx, type = 'ROTATION')
        if curve_node is not None:
            curve_data = get_curve_node_data(curve_node)
            if curve_data:
                fx_data['ROTATION_CURVE'] = curve_data    
    return fx_data

# Set the data of an effect
def set_fx_data(item, fx, fx_data):
    # Set the properties
    for key, value in fx_data.items():
        if key in ['ROTATION_CURVE', 'OFFSET_CURVE']:
            continue
        if not hasattr(fx, key):
            print(fx.type, ', Property Not Found:', key)
            continue
        if key == 'name':
            names = item.fx.keys()
            if value in names:
                value = get_prop_name('FX', names)
        try:
            setattr(fx, key, value)
        except Exception as e:
            print('Error while setting data:', e)
            continue
    # set offset curve data
    curve_data = fx_data.get('OFFSET_CURVE')
    if curve_data is not None:
        curve_node = get_curve_node(item, fx, type = 'EASING')
        if curve_node is not None:
            set_curve_node_data(curve_node, curve_data)
    # set rotation curve data
    curve_data = fx_data.get('ROTATION_CURVE')
    if curve_data is not None:
        curve_node = get_curve_node(item, fx, type = 'ROTATION')
        if curve_node is not None:
            set_curve_node_data(curve_node, curve_data)
    
            
# Get the data of an exposed property
def get_exposed_fx_data(item, exposed_fx):
    exp_data = {}
    for i in exposed_fx_prop:
        if not hasattr(exposed_fx, i):
            print('Attribute', i, 'Not found')
            continue
        value = getattr(exposed_fx, i)
        exp_data[i] = value
    return exp_data

# Set the data of an exposed property
def set_exposed_fx_data(exposed_fx, exp_data):
    for key, value in exp_data.items():
        if not hasattr(exposed_fx, key):
            print('Attribute', key, 'Not found')
            continue
        setattr(exposed_fx, key, value)
            
# Give the properties unique names
def prepare_fx_names(preset_data, item):
    names = item.fx.keys()
    # extract the numbers
    if not len(names):
        num = 0
    elif names == ['FX',]:
        num = 2
    else:
        nums = [int(i.replace('FX', '')) for i in names if i != 'FX']
        num = max(nums) + 1
    new_keys = {}
    effects = [preset_data[i].get('name') for i in preset_data if i != 'exposed_fx']
    for k in effects:
        n = (int(k[2:]) if k != 'FX' else 0) + num
        new_keys[k] = 'FX' if n == 0 else 'FX' + str(n)
    
    # Change the numbers
    # The Effects    
    for k, fx_data in preset_data.items():
        if k == 'exposed_fx':
            continue
        for key, value in fx_data.items():
            if key == "name" or (key == "sync_with" and value):
                if new_keys.get(value) == None:
                    continue
                fx_data[key] = new_keys[value]
    # exposed props
    e = preset_data.get('exposed_fx')
    if e is None:
        return preset_data
    new_porps = []
    for i in e:
        for ek, et in i.items():
            prop = {}
            prop[new_keys[ek]] = et
            new_porps.append(prop)
    preset_data['exposed_fx'] = new_porps
    return preset_data


# Duplicate The Selected Effect
def duplicate_fx(context, item):
    coll = item.fx
    index = item.fx_index
    new_index = index+1
    count = len(coll)
    fx = coll[index]
    new_fx = coll.add()
    
    # Set the properties
    fx_data = get_fx_data(item, fx)
    set_fx_data(item, new_fx, fx_data)
    
    new_fx.ui_name += ' (Copy)'
    item.fx_index = count
    coll.move(count, new_index)
    item.fx_index = new_index
    
# Copy Text item Properties
def copy_text_props(item, new_item):
    text_node = get_text_node(item)
    new_text_node = get_text_node(new_item)
    if not (text_node and new_text_node):
        print('Text node not found')
        return
    new_text_node.font = text_node.font
    new_text_node.overflow = text_node.overflow
    new_text_node.overflow = text_node.overflow
    new_text_node.align_x = text_node.align_x
    new_text_node.align_y = text_node.align_y
    new_text_node.pivot_mode = text_node.pivot_mode
    for idx, i in enumerate(text_node.inputs):
        new_text_node.inputs[idx].default_value = i.default_value
    # exturude
    extrude_node = get_extrude_node(item)
    new_extrude_node = get_extrude_node(new_item)
    if not (extrude_node and new_extrude_node):
        print('Extrude node not found')
        return
    extrude = extrude_node.inputs[1].default_value
    new_extrude_node.inputs[1].default_value = extrude
    
# Copy custom node group inputs
def copy_cusom_ng_inputs(item, new_item):
    ng = item_ng(item)
    new_ng = item_ng(new_item)
    if not(ng and new_ng):
        print('Node Group Not Found')
        return
    custom_ng = ng.nodes.get('ANIMAX_Custom_NG')
    new_custom_ng = new_ng.nodes.get('ANIMAX_Custom_NG')
    if not (custom_ng and new_custom_ng):
        print('Custom Node Group Not Found')
        return
    for idx, i in enumerate(custom_ng.inputs):
        if not hasattr(i, 'default_value'):
            continue
        value = getattr(i, 'default_value')
        setattr(new_custom_ng.inputs[idx], 'default_value', value)
    
# Copy item properties
def copy_item_props(context, item, new_item):
    # Copy the properties
    for prop in items_props:
        if not hasattr(item, prop):
            continue
        value = getattr(item, prop)
        setattr(new_item, prop, value)
    # copy visibility
    container = item.get('geo_node_container')
    new_container = new_item.get('geo_node_container')
    if container and new_container:
        new_container.matrix_world = container.matrix_world
        mod = container.modifiers.get('ANIMAX')
        new_mod = new_container.modifiers.get('ANIMAX')
        if mod and new_mod:
            new_mod.show_viewport = mod.show_viewport
            new_mod.show_render = mod.show_render
    # copy the material
    mat_node = get_material_node(item)
    new_mat_node = get_material_node(new_item)
    if mat_node and new_mat_node:
        mat = mat_node.inputs[2].default_value
        new_mat_node.inputs[2].default_value = mat
    # Copy Node Group Inputs
    if item.type == 'NODE_GROUP':
        copy_cusom_ng_inputs(item, new_item)
        
    if not item.type == 'TEXT':
        return
    # Copy text node
    copy_text_props(item, new_item)
    

# Duplicate The Selected Item
def duplicate_item(context):
    coll, index = items_coll(context)
    item = coll[index]
    new_index = index+1
    count = len(coll)
    new_item = add_item(context, item.type)
    new_item.ui_name = item.ui_name + ' (Copy)'
    # Copy the item data
    copy_item_props(context, item, new_item)    
    # Copy the effets
    for fx in item.fx:
        new_fx = new_item.fx.add()
        fx_data = get_fx_data(item, fx)
        set_fx_data(new_item, new_fx, fx_data)
    # Copy the exposed properties
    for exp_fx in item.exposed_fx:
        new_exp_fx = new_item.exposed_fx.add()
        exp_data = get_exposed_fx_data(item, exp_fx)
        set_exposed_fx_data(new_exp_fx, exp_data)
    # Move the new item
    coll.move(count, new_index)
    # Select the new item
    context.scene.animax_items.index = new_index
    #generate_items_node_tree(context, item)
    
# Delete exposed fx prop
def remove_exposed_fx_prop(item, fx):
    exposed_fx = item.exposed_fx
    for idx in reversed(range(len(exposed_fx))):
        if exposed_fx[idx].fx == fx.name:
            exposed_fx.remove(idx)
  
# generate the node-tree of the effects on property change
def update_fx_node_tree(self, context):
    scn = context.scene
    coll, index = items_coll(context)
    item = coll[index]
    generate_items_node_tree(context, item)
    return None

############################################################################################
################################## Animation Presets #######################################
############################################################################################

# prepare effects data for saving as json
def prepare_fx_data(fx_data):
    valid_types = [str, int, float, Vector, Euler, bool, list]
    for key, value in fx_data.items():
        if not type(value) in valid_types:
            fx_data[key] = None
            continue
        if hasattr(value, '__getitem__'):
            fx_data[key] = value[:]
    return fx_data

# Save a new preset
def save_fx_preset(item, file_path):
    fx_preset = {}
    # The effects
    for fx in item.fx:
        fx_data = get_fx_data(item, fx)
        fx_data = prepare_fx_data(fx_data)
        fx_preset[fx.name] = fx_data
    # Exposed props
    exposed_fx = []
    for i in item.exposed_fx:
        prop_data = {}
        prop_data[i.fx] = [i.id, i.ui_name, i.original_name, i.description]
        exposed_fx.append(prop_data)
    fx_preset['exposed_fx'] = exposed_fx
    # Write to json file
    with open(file_path, 'w') as json_file:
        json.dump(fx_preset, json_file, indent=5)
        
def read_preset_file(preset):
    with open(preset, 'r') as json_file:
        fx_preset = json.load(json_file)
    return fx_preset
    
        
# Load an animation preset
def load_fx_preset(item, preset):
#    with open(preset, 'r') as json_file:
#        fx_preset = json.load(json_file)
    fx_preset = read_preset_file(preset)
    fx_preset = prepare_fx_names(fx_preset, item)
    # Load the effects
    for key, fx_data in fx_preset.items():
        if key == 'exposed_fx':
            continue
        fx = item.fx.add()
        set_fx_data(item, fx, fx_data)
    # Load the exposet properties
    exposed_fx = fx_preset.get('exposed_fx')
    if exposed_fx is None:
        # resret the index
        item.fx_index = 0
        return
    for i in exposed_fx:
        exp_prop = item.exposed_fx.add()
        for k, v in i.items():
            exp_prop.fx = k
            exp_prop.id = v[0]
            exp_prop.ui_name = v[1]
            exp_prop.original_name = v[2]
            exp_prop.description = v[3]
    # resret the index
    item.fx_index = 0
    
# List of the animation Presets
def get_animation_presets():
    presets = []
    if not os.path.exists(anim_presets):
        return presets
    for p in os.listdir(anim_presets):
        if p.lower().endswith(".animax"):
            presets.append(p.replace(".animax", ""))
    return presets

# List of the custom animation Presets
def get_custom_animation_presets(context):
    presets = []
    custom_folder = custom_presets_dir(context)
    if custom_folder is None:
        return presets
    custom_anim_presets = os.path.join(custom_folder, 'Animation_Presets')
    if not os.path.exists(custom_anim_presets):
        return presets
    if os.path.samefile(custom_anim_presets, anim_presets):
        return presets
    for p in os.listdir(custom_anim_presets):
        if p.lower().endswith(".animax"):
            presets.append(p.replace(".animax", ""))
    return presets

# list of all the presets (custom and builtin)
def get_all_presets(context):
    presets_enum = []
    scn = context.scene
    filter = scn.animax_presets_filter
    a_presets = scn.animax_a_presets
    c_presets = scn.animax_c_presets
    filter = filter.strip().lower()
    presets = get_animation_presets()
    custom_presets = get_custom_animation_presets(context)
    a_icon = get_custom_icon_id('a_preset.png')
    c_icon = get_custom_icon_id('c_preset.png')
    if a_presets:
        presets = [i for i in presets if filter in i.lower()]
        for i, p in enumerate(presets):
            presets_enum.append(('A_'+p, p, '', a_icon, i))
            
    if c_presets:
        count = len(presets_enum)
        custom_presets = [i for i in custom_presets if filter in i.lower()]
        for i, p in enumerate(custom_presets):
            presets_enum.append(('C_'+p, p, '', c_icon, i+count))
    presets_enum = presets_enum if presets_enum else [('_Empty_', '_Empty_', '')]
    return presets_enum

# List of the animation Presets (Enum)
def animation_presets_enum(self, context):
    presets_enum = get_all_presets(context)
    return presets_enum

# Number of custom presets
def num_custom_presets(context):
    animation_presets = get_custom_animation_presets(context)
    slicing_presets = animax_slicing.get_custom_slicing_presets(context)
    return len(animation_presets), len(slicing_presets)

############################################################################################
################################### Clipboard Utils ########################################
############################################################################################
# Save to Clipboard
def save_clipboard(item, effects):
    clipboard = {}
    for fx in item.fx:
        if not fx.name in effects:
            continue
        fx_data = get_fx_data(item, fx)
        clipboard[fx.name] = fx_data
    with open(clipboard_path, 'w') as clip:
        clip.write(str(clipboard))

# Load Clipboard
def load_clipboard(item):
    clipboard = {}
    with open(clipboard_path, 'r') as clip:
        clipboard = eval(clip.read())
    if not type(clipboard) == dict:
        raise Exception('Wrong File Format')
        return
    clipboard = prepare_fx_names(clipboard, item)
    for fx_data in clipboard.values():
        fx = item.fx.add()
        set_fx_data(item, fx, fx_data)
    item.fx_index = 0
    
############################################################################################
#################################### Baking utils ##########################################
############################################################################################

# get the visiblity of the GN modifiers
def get_mod_vidibility(context):
    modifiers_visibility = []
    coll, index = items_coll(context)
    for i in coll:
        container = i.get('geo_node_container')
        if container is None:
            modifiers_visibility.append(None)
            continue
        mod = container.modifiers.get('ANIMAX')
        if mod is None:
            modifiers_visibility.append(None)
            continue
        modifiers_visibility.append(mod.show_viewport)
    return modifiers_visibility
            
# set the visiblity of the GN modifiers    
def set_mod_vidibility(context, mod_vis):
    coll, index = items_coll(context)
    if not len(coll) == len(mod_vis):
        return
    for i, c in enumerate(coll):
        if mod_vis[i] == None:
            continue
        container = c.get('geo_node_container')
        if container is None:
            continue
        mod = container.modifiers.get('ANIMAX')
        if mod is None:
            continue
        mod.show_viewport = mod_vis[i]

        
############################################################################################
###################################### Operators ###########################################
############################################################################################

# Load the animation properties
class ANIMAX_OT_load_props(Operator):
    bl_idname = "animax.load_props"
    bl_label = "Load"
    bl_options = {'UNDO'}
    bl_description = "Load ANIMAX properties to the blend file"
    
    def execute(self, context):
        # Animation Properties
        try:
            load_anim_props()
            items = context.scene.animax_items.coll.update()
        except Exception as e:
            self.report({'WARNING'}, 'Loading the properties failed: ' + str(e))
            return {'CANCELLED'}
        return {'FINISHED'}

# Add an item
class ANIMAX_OT_add_item(Operator):
    bl_idname = "animax.add_item"
    bl_label = "Add Item"
    bl_options = {'UNDO'}
    bl_description = "Add a new Item"
    
    item_type : StringProperty()
    
    @classmethod
    def poll(self, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        return True
    
    def execute(self, context):
        add_item(context, self.item_type)
        return {'FINISHED'}
    
# Remove item
class ANIMAX_OT_remove_item(Operator):
    bl_idname = "animax.remove_item"
    bl_label = "Remove"
    bl_options = {'UNDO'}
    bl_description = "Remove the selected Item"
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll = context.scene.animax_items.coll
        return len(coll)
    
    def execute(self, context):
        remove_item(context)
        return {'FINISHED'}
    
# Items Visibility
class ANIMAX_OT_items_visibility(Operator):
    bl_idname = "animax.items_visibility"
    bl_label = "Items Visibility"
    bl_options = {'UNDO'}
    bl_description = "Visibility of the items"
    
    action : StringProperty()
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll = context.scene.animax_items.coll
        return len(coll)
    
    def execute(self, context):
        coll, index = items_coll(context)
        item = coll[index]
        # attributs = {action: [attribute, value]}
        attributs = {
            'HIDE_VIEWPORT': ['show_viewport', False],
            'SHOW_VIEWPORT': ['show_viewport', True],
            'HIDE_RENDER': ['show_render', False],
            'SHOW_RENDER': ['show_render', True],
            'SOLO': ['show_viewport', False],
        }
        
        for i in coll:
            container = i.get('geo_node_container')
            if container is None:
                continue
            mod = container.modifiers.get('ANIMAX')
            if mod is None:
                continue
            attr, value = attributs[self.action]
            setattr(mod, attr, value)
            if self.action == 'SOLO' and i == item:
                mod.show_viewport = True
                    
        return {'FINISHED'}
    
# Add an effect
class ANIMAX_OT_add_fx(Operator):
    bl_idname = "animax.add_fx"
    bl_label = "Add Effect"
    bl_options = {'UNDO'}
    bl_description = "Add an effect"
    
    fx_type : StringProperty()
    
    @classmethod
    def poll(self, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        return len(scn.animax_items.coll)
    
    def execute(self, context):
        coll, index = items_coll(context)
        item = coll[index]
        fx_name = effects_dict[self.fx_type][0]
        add_fx(context, item, self.fx_type, fx_name)
        return {'FINISHED'}

# Remove effect    
class ANIMAX_OT_remove_fx(Operator):
    bl_idname = "animax.remove_fx"
    bl_label = "Remove"
    bl_options = {'UNDO'}
    bl_description = "Remove the selected Effect"
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        return len(coll[index].fx)
    
    def execute(self, context):
        coll, index = items_coll(context)
        item = coll[index]
        remove_fx(context, item)
        return {'FINISHED'}

# Duplicate effect    
class ANIMAX_OT_dplicate_fx(Operator):
    bl_idname = "animax.duplicate_fx"
    bl_label = "Duplicate"
    bl_options = {'UNDO'}
    bl_description = "Duplicate the selected Effect"
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        return len(coll[index].fx)
    
    def execute(self, context):
        coll, index = items_coll(context)
        item = coll[index]
        duplicate_fx(context, item)
        return {'FINISHED'}
    
# Duplicate item 
class ANIMAX_OT_dplicate_item(Operator):
    bl_idname = "animax.duplicate_item"
    bl_label = "Duplicate"
    bl_options = {'UNDO'}
    bl_description = "Duplicate the selected Item"
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        return len(coll)
    
    def execute(self, context):
        duplicate_item(context)
        return {'FINISHED'}
    
# Move the selected item
class ANIMAX_OT_move_item(Operator):
    bl_idname = "animax.move_item"
    bl_label = "Move Selected"
    bl_options = {'UNDO'}
    bl_description = "Move Selected Item"
    
    direction : StringProperty(name = 'Direction')
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        return len(coll) > 1
    
    def execute(self, context):
        scn = context.scene
        coll, index = items_coll(context)
        
        if self.direction == 'UP':
            if index > 0:
                coll.move(index, index-1)
                scn.animax_items.index -= 1
        elif self.direction == 'DOWN':
            if index < len(coll)-1:
                coll.move(index, index+1)
                scn.animax_items.index += 1
        return {'FINISHED'}
    
# Move the selected effect
class ANIMAX_OT_move_fx(Operator):
    bl_idname = "animax.move_fx"
    bl_label = "Move Selected"
    bl_options = {'UNDO'}
    bl_description = "Move Selected Effect"
    
    direction : StringProperty(name = 'Direction')
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        return len(coll[index].fx) > 1
    
    def execute(self, context):
        coll, index = items_coll(context)
        item = coll[index]
        fx_coll = item.fx
        fx_index = item.fx_index
        
        if self.direction == 'UP':
            if fx_index > 0:
                fx_coll.move(fx_index, fx_index-1)
                item.fx_index -= 1
                generate_items_node_tree(context, item)
        elif self.direction == 'DOWN':
            if fx_index < len(fx_coll)-1:
                fx_coll.move(fx_index, fx_index+1)
                item.fx_index += 1
                generate_items_node_tree(context, item)
        return {'FINISHED'}
    
# Move the selected property
class ANIMAX_OT_move_exposed_prop(Operator):
    bl_idname = "animax.move_exposed_prop"
    bl_label = "Move Selected"
    bl_options = {'UNDO'}
    bl_description = "Move this parameter"
    
    direction : StringProperty(name = 'Direction')
    index : IntProperty(name = 'Direction')
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        return len(coll[index].exposed_fx) > 1
    
    def execute(self, context):
        coll, index = items_coll(context)
        item = coll[index]
        exposed_fx = item.exposed_fx
        
        if self.direction == 'UP':
            if self.index > 0:
                exposed_fx.move(self.index, self.index-1)
                item.exposed_fx_index -= 1
        elif self.direction == 'DOWN':
            if self.index < len(exposed_fx)-1:
                exposed_fx.move(self.index, self.index+1)
                item.exposed_fx_index += 1
        return {'FINISHED'}
    
# Enable/Disable Items
class ANIMAX_OT_enable_items(Operator):
    bl_idname = "animax.enable_items"
    bl_label = "Enable/Disable Items"
    bl_options = {'UNDO'}
    bl_description = "Items Enable/Disable Options"
    
    action : StringProperty()
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll = context.scene.animax_items.coll
        return len(coll)
    
    def execute(self, context):
        coll, index = items_coll(context)
        item = coll[index]
        count = len(coll)
        
        seq = [True]*count if self.action == 'ENABLE' else [False]*count
        
        if self.action == 'SOLO':
            seq[index] = True
        coll.foreach_set('fx_enabled', seq)
        # not sure if necessary, for updating the UI List
        coll.update()
        # generate only once
        generate_items_node_tree(context, item)
        return {'FINISHED'}
    
# Enable/Disable Effects
class ANIMAX_OT_enable_fx(Operator):
    bl_idname = "animax.enable_fx"
    bl_label = "Enable/Disable Effects"
    bl_options = {'UNDO'}
    bl_description = "Effects Enable/Disable Options"
    
    action : StringProperty()
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        return len(coll[index].fx)
    
    def execute(self, context):
        coll, index = items_coll(context)
        item = coll[index]
        fx_coll = item.fx
        fx_index = item.fx_index
        count = len(fx_coll)
        
        seq = [True]*count if self.action == 'ENABLE' else [False]*count
        
        if self.action == 'SOLO':
            seq[fx_index] = True
        fx_coll.foreach_set('enabled', seq)
        # not sure if necessary, for updating the UI List
        coll.update()
        # generate only once
        generate_items_node_tree(context, item)
        return {'FINISHED'}
    
# Remove all the items
class ANIMAX_OT_remove_all_items(Operator):
    bl_idname = "animax.remove_all_items"
    bl_label = "Remove"
    bl_options = {'UNDO'}
    bl_description = "Remove all the Items"
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll = context.scene.animax_items.coll
        return len(coll)
    
    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        layout = self.layout
        layout.separator()
        message = 'Delete all the Items?'
        layout.label(text = message, icon = 'QUESTION')        
        layout.separator()
    
    def execute(self, context):
        len_coll = len(context.scene.animax_items.coll)
        for i in range(len_coll):
            context.scene.animax_items.index = 0
            remove_item(context)
        return {'FINISHED'}
    
# Remove all the effects
class ANIMAX_OT_remove_all_fx(Operator):
    bl_idname = "animax.remove_all_fx"
    bl_label = "Remove"
    bl_options = {'UNDO'}
    bl_description = "Remove all the Effects"
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        return len(coll[index].fx)
    
    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        layout = self.layout
        layout.separator()
        message = 'Delete all the Effects?'
        layout.label(text = message, icon = 'QUESTION')        
        layout.separator()
    
    def execute(self, context):
        coll, index = items_coll(context)
        item = coll[index]
        for fx in item.fx:
            delete_fx_node(item, fx)
            remove_exposed_fx_prop(item, fx)
        item.fx.clear()
        generate_items_node_tree(context, item)
        return {'FINISHED'}

# Load Font
class ANIMAX_OT_load_font(Operator, ImportHelper):
    bl_idname = "animax.load_font"
    bl_label = "Load Font"
    bl_options = {'UNDO'}
    bl_description = "Load a new font from a file (.ttf, .otf)"
    
    filepath : StringProperty(subtype="FILE_PATH")
    
    filter_glob : StringProperty(
        default='*.ttf;*.otf',
        options={'HIDDEN'}
    )
    
    def execute(self, context):
        coll, index = items_coll(context)
        if not len(coll):
            self.report({'WARNING'}, 'No Active Item!')
            return {'CANCELLED'}
        
        scn = context.scene
        fonts =  bpy.data.fonts
        filename, file_extension = os.path.splitext(self.filepath)
        font_name = os.path.basename(self.filepath)
        
        extensions = ['.ttf', '.otf']
        
        if file_extension.lower() not in extensions:
            self.report({'WARNING'}, 'The selected file is not (.ttf) or (.otf).')
            return {'CANCELLED'}
        
        item = coll[index]
            
        try:
            font = fonts.load(self.filepath)
        except Exception as e:
            self.report({'WARNING'}, str(e))
            return {'CANCELLED'}
        else:
            ng = item_ng(item)
            text_node = ng.nodes.get('ANIMAX_Text')
            if text_node is None:
                self.report({'WARNING'}, 'Text Node Not Found.')
                return {'CANCELLED'}
            text_node.font = font
            
        return {'FINISHED'}
    
    def invoke(self, context, event):
        if not self.filepath:
            font_dir = context.preferences.filepaths.font_directory
            if os.path.exists(font_dir):
                self.filepath = font_dir
        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}
    
# Load Image
class ANIMAX_OT_load_image(Operator):
    bl_idname = "animax.load_image"
    bl_label = "Load Image"
    bl_description = ("Load a new image.\n"
        "Hold Ctrl to open the textures folder")
    
    filepath : StringProperty(subtype="FILE_PATH")
        
    def execute(self, context):
        coll, index = items_coll(context)
        if not len(coll):
            self.report({'WARNING'}, 'No Active Item!')
            return {'CANCELLED'}
        item = coll[index]
        fx_coll = item.fx
        fx_index = item.fx_index
        if not len(fx_coll):
            self.report({'WARNING'}, 'No Active Effect!')
            return {'CANCELLED'}
        effect = fx_coll[fx_index]
        
        scn = context.scene
        image = None
        images =  bpy.data.images
        filename, file_extension = os.path.splitext(self.filepath)
        img_name = os.path.basename(self.filepath)
        
        extensions = ['.png', '.jpg', '.jpeg', '.bmp',
                      '.targa', '.avi', '.mp4', '.ogg',
                      '.flv', '.mov', '.mpeg', '.wmv',
                      '.exr', '.hdr', '.tif', '.webp']
        
        if file_extension.lower() not in extensions:
            self.report({'WARNING'}, 'The selected file is not a valid image file.')
            return {'CANCELLED'}
            
        try:
            image = images.load(self.filepath)
        except Exception as e:
            self.report({'WARNING'}, str(e))
            return {'CANCELLED'}
        else:
            effect.image_texture = image
            
        return {'FINISHED'}
    
    def invoke(self, context, event):
        if os.path.exists(textures_path):
            if not self.filepath or event.ctrl:
                self.filepath = textures_path + os.sep
        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}
    
# Copy name
class ANIMAX_OT_copy_name(Operator):
    bl_idname = "animax.copy_name"
    bl_label = "Copy Name"
    bl_options = {'UNDO'}
    bl_description = "Use this name as the name of the Item"
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        return len(coll)
    
    def execute(self, context):
        coll, index = items_coll(context)
        item = coll[index]
        
        if item.type == 'COLLECTION':
            if item.collection and hasattr(item.collection, 'name'):
                item.ui_name = item.collection.name
        if item.type == 'NODE_GROUP':
            if item.custom_ng and hasattr(item.custom_ng, 'name'):
                item.ui_name = item.custom_ng.name
        
        return {'FINISHED'}

######################################### Presets ##########################################
# Save the effects as a preset
class ANIMAX_OT_save_fx_preset(Operator):
    bl_idname = "animax.save_fx_preset"
    bl_label = "Save"
    bl_options = {'UNDO'}
    bl_description = "Save the effects as a preset"
    
    name : StringProperty(
        name = 'Name',
        default = 'My Preset',
        description = 'Name of the new preset',
    )
    presets = []
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        return len(coll[index].fx)
    
    def invoke(self, context, event):
        presets = get_custom_animation_presets(context)
        self.name = get_prop_name('My_Preset', presets)
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        layout = self.layout
        col = layout.column()
        
        custom_folder = custom_presets_dir(context)
        if custom_folder is None:
            preferences = prefs(context)
            col.label(text = 'You need to select the presets folder first', icon = 'INFO')
            col.label(text = 'Or you can set it up from the preferences', icon = 'BLANK1')
            col.prop(preferences, 'presets_path', text = '')
            return
        
        name = bpy.path.clean_name(self.name)
        presets = [i.lower() for i in get_custom_animation_presets(context)]
        col.alert = name.lower() in presets or not name.strip()
        col.label(text = 'Name of the new Preset')
        col.prop(self, 'name', text = '')
        if name.lower() in presets:
            col = layout.column()
            msg = 'The name "' + name + '" already exists!'
            col.label(text = msg, icon = 'ERROR')
            col.label(text = 'Press OK if you want to overide the preset.', icon = 'INFO')
    
    def execute(self, context):
        coll, index = items_coll(context)
        item = coll[index]
        name = bpy.path.clean_name(self.name)
        
        custom_folder = custom_presets_dir(context)
        if custom_folder is None:
            self.report({'WARNING'}, 'Custom presets folder not found')
            return {'CANCELLED'}
        custom_anim_presets = os.path.join(custom_folder, 'Animation_Presets')
        if not os.path.exists(custom_anim_presets):
            try:
                os.makedirs(custom_anim_presets)
            except Exception as e:
                self.report({'WARNING'}, 'Creating the presets folder failed: ' + str(e))
                return {'CANCELLED'}
        
        if not name.strip():
            self.report({'WARNING'}, 'Wrong name')
            return {'CANCELLED'}
        
        preset = os.path.join(custom_anim_presets, name + '.animax')
        save_fx_preset(item, preset)
        
        if os.path.exists(preset):
            self.report({'INFO'}, name + '.animax' + ' : Saved Successfully!')
        else:
            self.report({'WARNING'}, 'Something Went Wrong! Preset Not Saved')
        return {'FINISHED'}
    
# Load the effects as a preset
class ANIMAX_OT_load_fx_preset(Operator):
    bl_idname = "animax.load_fx_preset"
    bl_label = "Load"
    bl_options = {'UNDO'}
    bl_description = ("Load the selected Preset.\n"
        "Hold Ctrl to add to the effects")
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        return scn.animax_fx_presets != '_Empty_'
    
    def invoke(self, context, event):
        preset = context.scene.animax_fx_presets
        if preset == '_Empty_':
            self.report({'WARNING'}, 'Empty Presets Folders.')
            return {'CANCELLED'}
            
        coll, index = items_coll(context)
        item = coll[index]
        
        #delete all the effects
        if not event.ctrl and len(item.fx):
            for fx in item.fx:
                delete_fx_node(item, fx)
                remove_exposed_fx_prop(item, fx)
            item.fx_index = 0
            item.fx.clear()
        
        if preset.startswith('C_'):
            custom_folder = custom_presets_dir(context)
            presets_folder = os.path.join(custom_folder, 'Animation_Presets')
        elif preset.startswith('A_'):
            presets_folder = anim_presets
            
        preset = preset[2:]
        preset_file = os.path.join(presets_folder, preset+'.animax')
        if not os.path.exists(preset_file):
            self.report({'WARNING'}, "File doesn't exist")
            return {'CANCELLED'}
        
        try:
            load_fx_preset(item, preset_file)
        except Exception as e:
            self.report({'WARNING'}, str(e))
            return {'CANCELLED'}
        
        return {'FINISHED'}
    
# Delete an animation preset
class ANIMAX_OT_delete_fx_preset(Operator):
    bl_idname = "animax.delete_fx_preset"
    bl_label = "Delete"
    bl_options = {'UNDO'}
    bl_description = "Delete a preset"
    
    @classmethod
    def poll(cls, context):
        preset = context.scene.animax_fx_presets
        return preset is not None and preset != '_Empty_'
    
    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        preset = context.scene.animax_fx_presets
        preset = preset[2:]
        layout = self.layout
        layout.separator()
        message = 'Delete the preset: ' + preset + '?'
        layout.label(text = message, icon = 'QUESTION')        
        layout.separator()
    
    def execute(self, context):
        scn = context.scene
        preset = scn.animax_fx_presets
        if preset == '_Empty_':
            self.report({'WARNING'}, 'Nothing to delete.')
            return {'CANCELLED'}
        
        if preset.startswith('C_'):
            custom_folder = custom_presets_dir(context)
            presets_folder = os.path.join(custom_folder, 'Animation_Presets')
        elif preset.startswith('A_'):
            presets_folder = anim_presets
            
        preset = preset[2:]
        preset_file = os.path.join(presets_folder, preset+'.animax')
        if not os.path.exists(preset_file):
            self.report({'WARNING'}, "File doesn't exist")
            return {'CANCELLED'}
        
        try:
            os.remove(preset_file)
            presets = get_all_presets(context)
            if presets:
                scn.animax_fx_presets = presets[0][0]
        except Exception as e:
            self.report({'WARNING'}, str(e))
            return {'CANCELLED'}
        return {'FINISHED'}
    
######################################## Clipboard #########################################
# Save to Clipboard
class ANIMAX_OT_save_clipboard(Operator):
    bl_idname = "animax.save_clipboard"
    bl_label = "Copy"
    bl_description = "Copy the parameters of the effect(s)"
    
    options : EnumProperty(
        name = 'Options',
        items = (
            ('SELECTED','The Selected Effect',''),
            ('ENABLED','The Enabled Effects',''),
            ('ALL','All The Effects',''),
        ),
        default = 'SELECTED',
        description = 'Options',
    )
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        return len(coll[index].fx)
    
    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)
    
    def draw(self, context):
        layout = self.layout
        layout.separator()
        layout.prop(self, 'options')
        layout.separator()
        
    def execute(self, context):        
        coll, index = items_coll(context)
        item = coll[index]
        effects = []
        
        if self.options == 'SELECTED':
            fx_coll = item.fx
            fx_index = item.fx_index
            effects = [fx_coll[fx_index].name,]
        elif self.options == 'ENABLED':
            effects = [fx.name for fx in item.fx if fx.enabled]
        else:
            effects = [fx.name for fx in item.fx]
            
        if not effects:
            self.report({'WARNING'}, "Nothing to copy!")
            return {'CANCELLED'}
        
        save_clipboard(item, effects)
        return {'FINISHED'}
    
# Load The Clipboard
class ANIMAX_OT_load_clipboard(Operator):
    bl_idname = "animax.load_clipboard"
    bl_label = "Paste"
    bl_options = {'UNDO'}
    bl_description = "Paste the copied effect(s)"
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        if not os.path.exists(clipboard_path):
            return False
        return True
    
    def execute(self, context):
        coll, index = items_coll(context)
        item = coll[index]
        try:
            load_clipboard(item)
        except Exception as e:
            self.report({'WARNING'}, str(e))
            return {'CANCELLED'}
        
        return {'FINISHED'}
    
    
######################################## Opening Links #####################################
# Open the manual
class ANIMAX_OT_open_manual(Operator):
    bl_idname = "animax.open_manual"
    bl_label = "The Manual"
    bl_description = "Open The Manual"
        
    def execute(self, context):
        manual_dir = os.path.join(addon_folder, "Manual.pdf")
        if not os.path.exists(manual_dir):
            self.report({'WARNING'}, 'Manual file not found')
            return {'CANCELLED'}
        bpy.ops.wm.url_open(url = manual_dir)
        return {'FINISHED'}
    
# Open YT Playlist of tutorials
class ANIMAX_OT_yt_playlist(Operator):
    bl_idname = "animax.video_tuts"
    bl_label = "Video Tutorials"
    bl_description = "Open a playlist of video tutorials on YouTube"
        
    def execute(self, context):
        url = 'https://youtube.com/playlist?list=PLHQA94bJq141DtstuafgmmAUmT9L1RgQr'
        bpy.ops.wm.url_open(url = url)
        return {'FINISHED'}
    
############################################ Baking ########################################
    
# Bake to Shapekeys
class ANIMAX_OT_bake_to_shapekeys(Operator):
    bl_idname = "animax.bake_shapes"
    bl_label = "Bake to Shape Keys"
    bl_options = {'UNDO'}
    bl_description = "Convert the selected item to an animated object using Shape Keys"  
    
    
    start : IntProperty(
        default = 1, min = 0,
        max = 1000000,
        description = 'Start frame for baking'
    )
    end : IntProperty(
        default = 250,
        min = 1,
        max = 1000000,
        description = 'End frame for baking'
    )
    
    _timer = None
    realize_instances = True
    delete_invisible = True
    fx_enabled = True
    container = None
    gn_modifier = None
    obj = None
    collection = None
    new_collection = None
    vert_count = 0
    frame = 1
    frame_current = -1
    started = False
    finished = False
    merge_by_distance = False
    modifiers_visibility = []
                             
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        container = coll[index].get("geo_node_container")
        if container is None:
            return False
        return True
    
    def copy_uv_layers(self, eval_ob):
        obs = self.collection.all_objects
        # Available UV Layers
        uv_names = ['UVMap']
        for ob in obs:
            if not hasattr(ob.data, 'uv_layers'):
                continue
            for i in ob.data.uv_layers:
                if i.name in uv_names:
                    continue
                uv_names.append(i.name)
        # copy uvs
        for uv_name in uv_names:
            uv_map = self.obj.data.attributes.get(uv_name)
            if uv_map is None:
                continue
            # Make sure it's a 2D Vector
            if not uv_map.data_type == 'FLOAT2':
                continue
            count = len(uv_map.data)
            uvs = [0.0,]*count*2
            uv_map.data.foreach_get('vector', uvs)
            self.obj.data.attributes.remove(uv_map)
            uv_layer = self.obj.data.uv_layers.new(name=uv_name, do_init=False)
            uv_layer.data.foreach_set('uv', uvs)
        # remove FX attributes
        for atr in reversed(self.obj.data.attributes):
            if atr.name.startswith('FX'):
                self.obj.data.attributes.remove(atr)
    
    def bake_current_frame(self, context):
        if self.vert_count == 0:
            raise Exception("No vertices found")
        scn = context.scene
        self.frame_current = scn.frame_current
        depsgraph = context.evaluated_depsgraph_get()
        eval_ob = self.container.evaluated_get(depsgraph)
        eval_me = eval_ob.to_mesh()
        if len(eval_me.vertices) != self.vert_count:
            raise Exception("The number of vertices changed during the baking process")
        # get the coordinates of all the vertices at once
        coords = [0.0]*self.vert_count*3
        eval_me.vertices.foreach_get('co', coords)
        # add a new shape key
        shape_key_name = 'Frame ' + str(self.frame_current)
        shape_key = self.obj.shape_key_add(name = shape_key_name, from_mix = False)
        # before starting, disable relative for the shape keys, and copy the UVMap
        if self.start == self.frame_current:
            self.obj.data.shape_keys.use_relative = False
        # set the coordinates of all the vertices at once
        shape_key.data.foreach_set('co', coords)
        #if the baking finished
        if self.end == self.frame_current:
            # keyframe the eval_time of the shape keys
            shape_keys = self.obj.data.shape_keys
            shape_keys.eval_time = 0.0
            shape_keys.keyframe_insert('eval_time', frame = self.start)
            shape_keys.eval_time = 10.0 * (self.end - self.start)
            shape_keys.keyframe_insert('eval_time', frame = self.end)
            # change the interpolation to linear
            for fc in shape_keys.animation_data.action.fcurves:
                for pt in fc.keyframe_points:
                    pt.interpolation = 'LINEAR'
            self.finished = True
        # next frame
        if not self.finished:
            scn.frame_set(self.frame_current+1)
    
    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)
    
    def draw(self, context):
        layout = self.layout
        col = layout.column()
        col.separator()
        box = col.box()
        box.label(text = 'You can press Escap to cancel', icon = 'INFO')
        col.separator()
        col.prop(self, "start", text = 'Start Frame')
        col.prop(self, "end", text = 'End Frame')
        col.separator()
        
    def modal(self, context, event):
        # cancel on Escap key press    
        if event.type in {'ESC',}:
            self.cancel(context)
            self.report({'INFO'}, 'Baking Canceled!')
            return {'CANCELLED'}
        
        if event.type == 'TIMER':
            scn = context.scene
            # if the baking finished
            if self.finished:
                self.cancel(context)
                self.report({'INFO'}, 'Baking Finished!')
                return {'FINISHED'}
            # if the previous frame finished, bake the next frame
            if self.frame_current < scn.frame_current:
                try:
                    self.bake_current_frame(context)
                except Exception as e:
                    self.cancel(context)
                    self.report({'WARNING'}, str(e))
                    return {'CANCELLED'}
        return {'RUNNING_MODAL'}
        
    def execute(self, context):
        coll, index = items_coll(context)
        item = coll[index]
        container = item.get("geo_node_container")
        collections = bpy.data.collections
        scn = context.scene
        self.frame = scn.frame_current
        self.realize_instances = item.realize_instances
        self.delete_invisible = item.delete_invisible
        
        if self.start > self.end:
            self.report({'WARNING'}, "The End frame must be greater than the Start frame")
            return {'CANCELLED'}
        # check the geometry node container
        if container is None:
            self.report({'WARNING'}, "Container object not found")
            return {'CANCELLED'}
        self.container = container
        # check the modifier
        mod = container.modifiers.get('ANIMAX')
        if mod is None:
            self.report({'WARNING'}, "Geometry Nodes modifier not found")
            return {'CANCELLED'}
        # store the visibility of the items
        self.modifiers_visibility = get_mod_vidibility(context)
        set_mod_vidibility(context, [False]*len(coll))
        mod.show_viewport = True
        self.gn_modifier = mod
        # add a new collection
        new_collection = collections.new(item.ui_name + "_Baked")
        scn.collection.children.link(new_collection)
        self.new_collection = new_collection
        self.collection = item.collection
        # make sure the effects are enabled and realize instances is enabled 
        self.fx_enabled = item.fx_enabled
        item.fx_enabled = True        
        item.realize_instances = True
        item.delete_invisible = False
        if hasattr(item, 'merge_by_distance'):
            self.merge_by_distance = item.merge_by_distance
            item.merge_by_distance = False
        # set the start frame
        scn.frame_set(self.start)
        self.frame_current = self.start - 1
        # create the new object
        container.hide_viewport = False
        depsgraph = context.evaluated_depsgraph_get()
        container_eval = container.evaluated_get(depsgraph)
        me = bpy.data.meshes.new_from_object(container_eval)
        vert_count = len(me.vertices)
        self.vert_count = vert_count
        obj = bpy.data.objects.new(item.ui_name, me)
        obj.matrix_world = container.matrix_world
        new_collection.objects.link(obj)
        self.obj = obj
        # start
        wm = context.window_manager
        self._timer = wm.event_timer_add(0.1, window=context.window)
        wm.modal_handler_add(self)
        self.started = True
        return {'RUNNING_MODAL'}
    
    def cancel(self, context):
        # remove the timer
        if self._timer is not None:
            wm = context.window_manager
            wm.event_timer_remove(self._timer)
        if not self.started:
            return
        # restore the parameters
        scn = context.scene
        scn.frame_set(self.frame)
        coll, index = items_coll(context)
        item = coll[index]
        item.fx_enabled = self.fx_enabled
        item.realize_instances = self.realize_instances
        item.delete_invisible = self.delete_invisible
        if hasattr(item, 'merge_by_distance'):
            item.merge_by_distance = self.merge_by_distance
        # restore the visibility of the items
        set_mod_vidibility(context, self.modifiers_visibility)
        # Copy UVs and delete if necessary
        if self.finished:
            # Copy the UVs
            self.copy_uv_layers(self)
            return
        if self.obj:
            bpy.data.meshes.remove(self.obj.data)
        if self.new_collection:
            bpy.data.collections.remove(self.new_collection)
            
# Bake to Keyframes
class ANIMAX_OT_bake_to_keyframes(Operator):
    bl_idname = "animax.bake_keyframes"
    bl_label = "Bake to Keyframes"
    bl_options = {'UNDO'}
    bl_description = ("Transfer the procedural animation to the source objects using keyframes.\n"
        "Works only with the items of the type Collection")
    
    start : IntProperty(
        default = 1, min = 0,
        max = 1000000,
        description = 'Start frame for baking'
    )
    end : IntProperty(
        default = 250,
        min = 1,
        max = 1000000,
        description = 'End frame for baking'
    )
    
    _timer = None
    realize_instances = True
    delete_invisible = True
    fx_enabled = True
    container = None
    gn_modifier = None
    frame = 1
    frame_current = -1
    started = False
    finished = False
    collection = None
    material = None
    initial_transform = {}
    modifiers_visibility = []
    
                             
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        container = coll[index].get("geo_node_container")
        if container is None:
            return False
        if coll[index].type != 'COLLECTION':
            return False
        return True
    
    def keyframe_transform(self):
        for ob in self.collection.all_objects:
            ob.keyframe_insert('location')
            ob.keyframe_insert('rotation_euler')
            ob.keyframe_insert('scale')
    
    def bake_current_frame(self, context):
        scn = context.scene
        self.frame_current = scn.frame_current
        # keyframe the objects in the next frame, to avoid the extrapolation
        if scn.frame_current != self.end:
            self.frame_current = scn.frame_current+1
            self.gn_modifier.show_viewport = False
            scn.frame_set(self.frame_current)
            self.keyframe_transform()
            self.frame_current -= 1
            self.gn_modifier.show_viewport = True
            scn.frame_set(self.frame_current)
        # copy the matrix_world from the instances and keyframe
        depsgraph = context.evaluated_depsgraph_get()
        for i in depsgraph.object_instances:
            if not i.is_instance:
                continue
            if not i.parent.name == self.container.name:
                continue
            if i.object.name == self.container.name:
                continue
            src_ob = scn.objects.get(i.object.name)
            src_ob.matrix_world = i.matrix_world
            src_ob.keyframe_insert('location')
            src_ob.keyframe_insert('rotation_euler')
            src_ob.keyframe_insert('scale')
        #if the baking finished
        if self.end == self.frame_current:
            self.finished = True
        # next frame
        if not self.finished:
            scn.frame_set(self.frame_current+1)
    
    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)
    
    def draw(self, context):
        layout = self.layout
        col = layout.column()
        col.separator()
        box = col.box()
        box.label(text = 'You can press Escap to cancel', icon = 'INFO')
        col.separator()
        col.prop(self, "start", text = 'Start Frame')
        col.prop(self, "end", text = 'End Frame')
        col.separator()
        
    def modal(self, context, event):
        # cancel on Escap key press    
        if event.type in {'ESC',}:
            self.cancel(context)
            return {'CANCELLED'}
        
        if event.type == 'TIMER':
            scn = context.scene
            # if the baking finished
            if self.finished:
                self.cancel(context)
                self.report({'INFO'}, 'Baking Finished!')
                return {'FINISHED'}
            # if the previous frame finished, bake the next frame
            if self.frame_current < scn.frame_current:
                try:
                    self.bake_current_frame(context)
                except Exception as e:
                    self.cancel(context)
                    self.report({'WARNING'}, str(e))
                    return {'CANCELLED'}
        return {'RUNNING_MODAL'}
        
    def execute(self, context):
        coll, index = items_coll(context)
        item = coll[index]
        if item.type != 'COLLECTION':
            self.report({'WARNING'}, "Only Items of the type collection are supported")
            return {'CANCELLED'}
        container = item.get("geo_node_container")
        collections = bpy.data.collections
        scn = context.scene
        self.frame = scn.frame_current
        self.realize_instances = item.realize_instances
        self.delete_invisible = item.delete_invisible
        
        if item.collection is None:
            self.report({'WARNING'}, "No Collection Found")
            return {'CANCELLED'}
        
        if not hasattr(item.collection, 'all_objects'):
            self.report({'WARNING'}, "No Objects Found")
            return {'CANCELLED'}
        
        if self.start > self.end:
            self.report({'WARNING'}, "The End frame must be greater than the Start frame")
            return {'CANCELLED'}
        # check the geometry node container
        if container is None:
            self.report({'WARNING'}, "Container object not found")
            return {'CANCELLED'}
        self.container = container
        # check the modifier
        mod = container.modifiers.get('ANIMAX')
        if mod is None:
            self.report({'WARNING'}, "Geometry Nodes modifier not found")
            return {'CANCELLED'}
        # store the visibility of the items
        self.modifiers_visibility = get_mod_vidibility(context)
        set_mod_vidibility(context, [False]*len(coll))
        mod.show_viewport = True
        self.gn_modifier = mod
        # store the material, if there is a material set it will give wrong baking result
        mat_node = get_material_node(item)
        if mat_node is not None:
            self.material = mat_node.inputs[2].default_value
            mat_node.inputs[2].default_value = None
        # make sure the effects are enebled and realize instances is disabled
        self.fx_enabled = item.fx_enabled
        item.fx_enabled = True        
        item.realize_instances = False
        item.delete_invisible = False
        # set the start frame
        scn.frame_set(self.start)
        self.frame_current = self.start - 1
        container.hide_viewport = False
        # keyframe the objects of the collection in the first frame
        self.collection = item.collection
        self.keyframe_transform()
        # Clear the "FX" outputs, necessary for Blender 4.1
        node_group = mod.node_group
        items_tree = node_group.interface.items_tree
        sockets = [i for i in items_tree if i.item_type == 'SOCKET' and i.in_out == 'OUTPUT' and 'FX' in i.name]
        for socket in sockets:
            node_group.interface.remove(socket)
        # start
        wm = context.window_manager
        self._timer = wm.event_timer_add(0.1, window=context.window)
        wm.modal_handler_add(self)
        self.started = True
        return {'RUNNING_MODAL'}
    
    def cancel(self, context):
        # remove the timer
        if self._timer is not None:
            wm = context.window_manager
            wm.event_timer_remove(self._timer)
        if not self.started:
            return
        # restore the parameters
        scn = context.scene
        scn.frame_set(self.frame)
        coll, index = items_coll(context)
        item = coll[index]
        item.fx_enabled = self.fx_enabled
        item.realize_instances = self.realize_instances
        item.delete_invisible = self.delete_invisible
        # restore the material
        mat_node = get_material_node(item)
        if mat_node is not None:
            mat_node.inputs[2].default_value = self.material
        # restore the visibility of the items
        set_mod_vidibility(context, self.modifiers_visibility)
        if self.finished:
            self.gn_modifier.show_viewport = False
            self.gn_modifier.show_render = False
            
            
##################################### Right click menu #####################################
# Expose an effect's property in the items panel
class ANIMAX_OT_expose_fx_prop(Operator):
    bl_idname = "animax.expose_fx_prop"
    bl_label = "Expose this parameter"
    bl_options = {'UNDO'}
    bl_description = "Add this parameter to the Items panel"
    
    id : StringProperty()
    prop_name : StringProperty()
    ui_name : StringProperty()
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        return len(coll[index].fx)
    
    def execute(self, context):
        coll, index = items_coll(context)
        exposed_fx = coll[index].exposed_fx.add()
        exposed_fx.fx = self.id
        exposed_fx.id = self.prop_name
        exposed_fx.ui_name = self.ui_name
        exposed_fx.original_name = self.ui_name
        coll[index].exposed_fx.update()
        return {'FINISHED'}
    
# Remove an exposed property
class ANIMAX_OT_remove_exposed_prop(Operator):
    bl_idname = "animax.remove_exposed_prop"
    bl_label = "Remove"
    bl_options = {'UNDO'}
    bl_description = "Remove this parameter"
    
    index : IntProperty()
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        return len(coll[index].exposed_fx)
    
    def execute(self, context):
        coll, index = items_coll(context)
        exposed_fx = coll[index].exposed_fx
        exposed_fx.remove(self.index)
        exposed_fx.update()
        return {'FINISHED'}
    
# Exposed property info
class ANIMAX_OT_exposed_prop_info(Operator):
    bl_idname = "animax.exposed_prop_info"
    bl_label = "Info"
    bl_options = {'UNDO'}
    bl_description = "Informations about this parameter"
    
    index : IntProperty()
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        return len(coll)
    
    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)
    
    def draw(self, context):
        layout = self.layout
        coll, index = items_coll(context)
        item = coll[index]
        exposed_fx = item.exposed_fx
        prop = exposed_fx[self.index]
        fx = item.fx.get(prop.fx)
        layout.prop(prop, 'ui_name', text = 'UI Name')
        layout.label(text = "Effect ID: " + prop.fx)
        layout.label(text = "Effect Name: " + fx.ui_name)
        layout.label(text = "Effect Type: " + effects_dict[fx.type][0])
        layout.label(text = "Parameter's Name: " + prop.original_name)
        layout.prop(prop, 'description', text = '', icon = 'INFO')
        
    def execute(self, context):
        return {'FINISHED'}
    
# Exposed properties options
class ANIMAX_OT_exposed_prop_options(Operator):
    bl_idname = "animax.exposed_prop_options"
    bl_label = "Options"
    bl_options = {'UNDO'}
    bl_description = "Options of the exposed parameters"
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        return len(coll[index].exposed_fx)
    
    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)
    
    def draw(self, context):
        layout = self.layout
        preferences = prefs(context)
        layout.prop(preferences, 'remove_button')
        layout.prop(preferences, 'move_button')
        layout.prop(preferences, 'info_button')
        
    def execute(self, context):
        return {'FINISHED'}
    
######################################## Check Mesh ########################################
# Check the mesh before fracturing
class ANIMAX_OT_check_mesh(Operator):
    bl_idname = "animax.check_mesh"
    bl_label = "Check Object"
    bl_options = {'UNDO'}
    bl_description = "Check if the object is fracturable without issues"
    
    info = dict()
    elements = dict()
    verts = 0
    select : BoolProperty(
        name = 'Select',
        default = True,
        description = 'Select the mesh areas containing the issues'
    )
    
    @classmethod
    def poll(cls, context):
        ob = context.object
        if ob is None:
            return False
        if ob.mode != 'OBJECT':
            return False
        if ob.type != 'MESH':
            return False
        return ob in context.editable_objects
    
    def invoke(self, context, event):
        ob = context.object
        if len(ob.modifiers):
            depsgraph = context.evaluated_depsgraph_get()
            obj_eval = ob.evaluated_get(depsgraph)
            mesh = obj_eval.data
        else:
            mesh = ob.data
            for v in mesh.vertices:
                v.select = False
            mesh.update()
            context.view_layer.objects.active = ob
        self.verts = len(mesh.vertices)
        self.info, self.elements = check_mesh(mesh)
        return context.window_manager.invoke_props_dialog(self)
    
    def draw(self, context):
        ob = context.object
        layout = self.layout
        if not self.verts:
            id = get_custom_icon_id('error.png')
            layout.label(text = 'Empty Mesh', icon_value = id)
            return
        issues = [v for v in self.info.values() if v]
        icon = 'error.png' if len(issues) else 'ok.png'
        id = get_custom_icon_id(icon)
        layout.label(text = f'Issues Found: ({len(issues)})', icon_value = id)
        layout.separator()
        for k, v in self.info.items():
            if v: layout.label(text = f'{k}: {v}')
        if len(issues) and not len(ob.modifiers):
            layout.prop(self, 'select')
        
    def execute(self, context):
        if not self.select or not self.verts:
            return {'FINISHED'}
        #select the mesh elements
        ob = context.object
        if not len(ob.modifiers):
            bpy.ops.object.mode_set(mode='EDIT')
            bpy.ops.mesh.reveal()
            bpy.ops.mesh.select_all(action='DESELECT')
            bpy.ops.mesh.select_mode(type='VERT')
            bpy.ops.object.mode_set(mode='OBJECT')
            for k, v in self.elements.items():
                element = getattr(ob.data, k)
                for i in v:
                    element[i].select = True
        return {'FINISHED'}
    
##################################### Load Node Group ######################################
# Load the node-groups of the generators
class ANIMAX_OT_load_generators_ng(Operator):
    bl_idname = "animax.load_generators_ng"
    bl_label = "Load the generators"
    bl_options = {'UNDO'}
    bl_description = "Load the node-groups of the generators"
    
    @classmethod
    def poll(cls, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        item = coll[index]
        return item.type == 'NODE_GROUP'
    
    def execute(self, context):
        if not os.path.exists(generators_dir):
            self.report({'WARNING'}, "Blend file of the generators not found")
            return {'CANCELLED'}
        
        node_groups = bpy.data.node_groups
        loaded = {i.get('ANIMAX_GENERATOR') for i in node_groups if i.type == 'GEOMETRY'}
        names = ['Grid', 'Circular Array', 'Voxels', 'Scatter']
        generators = [i for i in names if i not in loaded]
        if not generators:
            return {'FINISHED'}
        
        libs = bpy.data.libraries    
        with libs.load(generators_dir, link=False) as (data_src, data_dst):
                data_dst.node_groups = generators
        return {'FINISHED'}

############################################################################################
####################################### The UI #############################################
############################################################################################

################################# Right click menu #########################################
# add an item to the right click menu
def draw_menu(self, context):
    #print(context.button_pointer.bl_rna.name)
    source_prop = None
    try:
        source_prop = context.button_pointer.bl_rna.name
        if not source_prop == 'AnimaxFX':
            return
        id = context.button_pointer.name
        prop_name = context.button_prop.identifier
        ui_name = context.button_prop.name
    except Exception as e:
        #print(str(e))
        return
    coll, index = items_coll(context)
    for i in coll[index].exposed_fx:
        if i.fx == id and i.id == prop_name:
            return
        
    layout = self.layout
    layout.separator()
    op = layout.operator("animax.expose_fx_prop", icon = 'EVENT_A')
    op.id = id
    op.prop_name = prop_name
    op.ui_name = ui_name

###################################### FX UI ###############################################
# Template for some panels Having the same attributs
class TemplateItems:
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "ANIMAX"
    
    @classmethod
    def poll(self, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        return scn.animax_tabs == 'ANIMATE'

# Template for some panels Having the same poll, same attributs
class TemplateFX:
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "ANIMAX"
    
    @classmethod
    def poll(self, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        item = coll[index]
        if not is_container_valid(item):
            return False
        fx = item.fx
        fx_index = item.fx_index
        if not len(fx):
            return False
        effect = fx[fx_index]
        return scn.animax_tabs == 'ANIMATE'

# Items Menu
class AnimaxItems(Menu):
    bl_idname = "ANIMAX_MT_items"
    bl_label = "Items"
    bl_description = "Add an Item"        

    def draw(self, context):
        scn = context.scene
        layout = self.layout
        for i in items_dict:
            name = items_dict[i][0]
            icon = items_dict[i][1]
            layout.operator("animax.add_item", text = name, icon = icon).item_type = i

# Effects Menu
class AnimaxEffects(Menu):
    bl_idname = "ANIMAX_MT_effects"
    bl_label = "Effects"
    bl_description = "Add an effect"        

    def draw(self, context):
        scn = context.scene
        layout = self.layout
        
        row = layout.row()
        
        col = row.column()
        col.label(text = 'Regular')
        col.separator()
        ico = get_custom_icon_id('transform.png')
        col.operator("animax.add_fx", text = 'Transform', icon_value = ico).fx_type = 'TRANSFORM'
        ico = get_custom_icon_id('explode.png')
        col.operator("animax.add_fx", text = 'Explode', icon_value = ico).fx_type = 'EXPLODE'
        ico = get_custom_icon_id('collapse.png')
        col.operator("animax.add_fx", text = 'Collapse', icon_value = ico).fx_type = 'COLLAPSE'
        ico = get_custom_icon_id('robotic.png')
        col.operator("animax.add_fx", text = 'Robotic', icon_value = ico).fx_type = 'ROBOTIC'
        ico = get_custom_icon_id('empty.png')
        col.operator("animax.add_fx", text = 'Empty', icon_value = ico).fx_type = 'EMPTY'
        
        col = row.column()
        col.label(text = 'Dependent')
        col.separator()
        ico = get_custom_icon_id('attract.png')
        col.operator("animax.add_fx", text = 'Attract', icon_value = ico).fx_type = 'ATTRACT'
        ico = get_custom_icon_id('repulse.png')
        col.operator("animax.add_fx", text = 'Repulse', icon_value = ico).fx_type = 'REPULSE'
        ico = get_custom_icon_id('follow_curve.png')
        col.operator("animax.add_fx", text = 'Follow Curve', icon_value = ico).fx_type = 'FOLLOW_CURVE'
        ico = get_custom_icon_id('orbit.png')
        col.operator("animax.add_fx", text = 'Orbit', icon_value = ico).fx_type = 'ORBIT'
        ico = get_custom_icon_id('project.png')
        col.operator("animax.add_fx", text = 'Project', icon_value = ico).fx_type = 'PROJECT'
        
        col = row.column()
        col.label(text = 'Generators')
        col.separator()
        ico = get_custom_icon_id('noise.png')
        col.operator("animax.add_fx", text = 'Noise', icon_value = ico).fx_type = 'NOISE'
        ico = get_custom_icon_id('wave.png')
        col.operator("animax.add_fx", text = 'Wave', icon_value = ico).fx_type = 'WAVE'
        
#        for i in effects_dict:
#            name = effects_dict[i][0]
#            layout.operator("animax.add_fx", text = name).fx_type = i
            
# Items Options
class Animax_Items_Options(Menu):
    bl_idname = "ANIMAX_MT_items_options"
    bl_label = "Options"
    bl_description = "Items Options"

    def draw(self, context):
        layout = self.layout
        layout.label(text='Options', icon = 'TOOL_SETTINGS')
        #layout.separator()
        #layout.operator("flares_wizard.duplicate_element", text='Duplicate', icon = 'DUPLICATE')
        layout.separator()
        layout.operator("animax.items_visibility", text='Show all in the viewport', icon = 'RESTRICT_VIEW_OFF').action = 'SHOW_VIEWPORT'
        layout.operator("animax.items_visibility", text='Hide all in the viewport', icon = 'RESTRICT_VIEW_ON').action = 'HIDE_VIEWPORT'
        layout.operator("animax.items_visibility", text='Show all in the render', icon = 'RESTRICT_RENDER_OFF').action = 'SHOW_RENDER'
        layout.operator("animax.items_visibility", text='Hide all in the render', icon = 'RESTRICT_RENDER_ON').action = 'HIDE_RENDER'
        layout.operator("animax.items_visibility", text='Solo the selected item', icon = 'VIS_SEL_11').action = 'SOLO'
        layout.separator()
        layout.operator("animax.enable_items", text='Enable All', icon = 'CHECKBOX_HLT').action = 'ENABLE'
        layout.operator("animax.enable_items", text='Disable All', icon = 'CHECKBOX_DEHLT').action = 'DISABLE'
        layout.operator("animax.enable_items", text='Solo Selected', icon = 'RESTRICT_SELECT_OFF').action = 'SOLO'
        layout.separator()
        layout.operator("animax.duplicate_item", icon = 'DUPLICATE')
        layout.separator()
        layout.operator("animax.bake_keyframes", icon = 'KEYINGSET')
        layout.operator("animax.bake_shapes", icon = 'SHAPEKEY_DATA')
        layout.separator()
        layout.operator("animax.remove_all_items", text='Delete all', icon = 'TRASH')
            
# Effects Options
class Animax_FX_Options(Menu):
    bl_idname = "ANIMAX_MT_fx_options"
    bl_label = "Options"
    bl_description = "Effects Options"

    def draw(self, context):
        layout = self.layout
        layout.label(text='Options', icon = 'TOOL_SETTINGS')
        layout.separator()
        layout.operator("animax.duplicate_fx", text='Duplicate', icon = 'DUPLICATE')
        layout.separator()
        layout.operator("animax.save_clipboard", icon = 'COPYDOWN')
        layout.operator("animax.load_clipboard", icon = 'PASTEDOWN')
        layout.separator()
        layout.operator("animax.save_fx_preset", text='Save', icon = 'EXPORT')
        layout.separator()
        layout.operator("animax.enable_fx", text='Enable All', icon = 'CHECKBOX_HLT').action = 'ENABLE'
        layout.operator("animax.enable_fx", text='Disable All', icon = 'CHECKBOX_DEHLT').action = 'DISABLE'
        layout.operator("animax.enable_fx", text='Solo Selected', icon = 'RESTRICT_SELECT_OFF').action = 'SOLO'
        layout.separator()
        layout.operator("animax.remove_all_fx", text='Delete all', icon = 'TRASH')
        
# UI list for the items
class SCENE_UL_animax_items(UIList):
   
   def draw_item(self, context, layout, data, item, icon, active_data, active_propname):
       row = layout.row()       
       ico = items_dict[item.type][1]
       row.prop(item, "ui_name", text="", emboss=False, icon = ico)
       row = layout.row(align = True)
       row.alignment = 'RIGHT'
       container = item.get('geo_node_container')
       
       if not is_container_valid(item):
           row.alert = True
           row.label(text = '', icon = 'ERROR')
           return
       
       try:
           mod = container.modifiers.get('ANIMAX')
           row.prop(mod, "show_viewport", text="", emboss=False)
           row.prop(mod, "show_render", text="", emboss=False)
       except:
           row.alert = True
           row.label(text = '', icon = 'ERROR')
           return
       
       row.prop(item, "fx_enabled", text="")
       
# UI list for the effects
class SCENE_UL_animax_fx(UIList):
   
   def draw_item(self, context, layout, data, item, icon, active_data, active_propname):
       coll, index = items_coll(context)
       fx = coll[index].fx
       ico = 'TEMP' if item.sync else 'BLANK1'
       row = layout.row(align = True)
       fx_icon = get_custom_icon_id(item.type.lower()+'.png')
       row.label(text = '', icon_value = fx_icon)
       if item.sync and get_item_idx(fx, item.name) != 0:
           row.prop(item, "ui_name", text="", emboss=False, icon = 'TEMP')
       else:
           row.prop(item, "ui_name", text="", emboss=False)
       row = layout.row()
       row.alignment = 'RIGHT'
       row.enabled = False
       row.scale_x = 0.6
       row.prop(item, "name", text="", emboss=False)
       row = layout.row()
       row.prop(item, "enabled", text="")

# ANIMAX Panel
class ANIMAX_PT_main(Panel):
    bl_label = "ANIMAX " + str(bl_info['version'])
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "ANIMAX"
        
    def draw(self, context):
        layout = self.layout
        if bpy.app.version < (4,1,0):
            box = layout.box()
            box.alert = True
            box.label(text = "Not compatible with Blender " + str(bpy.app.version), icon = 'ERROR')
            box.label(text = "Compatible with Blender (4,1,0) and above", icon = 'BLANK1')
            
        paths = context.preferences.filepaths
        auto_exec = paths.use_scripts_auto_execute
        
        if not auto_exec:
            box = layout.box()
            box.alert = True
            box.label(text = "Auto Run Python Scripts is disabled", icon = 'ERROR')
            box.label(text = "It's necessary for some functionalities", icon = 'BLANK1')
            box.prop(paths, 'use_scripts_auto_execute', icon = 'BLANK1')
            
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            layout.operator("animax.load_props", icon='FILE_SCRIPT')
            return
        
        if not hasattr(scn, 'animax_exposed_fx'):
            layout.operator("animax.load_props", icon='FILE_SCRIPT')
            return
        row = layout.row()
        row.prop(scn, "animax_tabs", expand = True)
        if scn.animax_tabs == 'FRACTURE':
            layout.operator("animax.check_mesh")
            layout.prop(scn, "animax_fracture_tabs", expand = True)
        
            
# Items Panel
class ANIMAX_PT_items(TemplateItems, Panel):
    bl_label = "Items"
    bl_parent_id = "ANIMAX_PT_main"
    
    @classmethod
    def poll(self, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        if not hasattr(scn, 'animax_exposed_fx'):
            return False
        return scn.animax_tabs == 'ANIMATE'
        
    def draw(self, context):
        scn = context.scene
        layout = self.layout
        coll, index = items_coll(context)
        
        row = layout.row()
        row.template_list("SCENE_UL_animax_items", "coll", scn.animax_items, "coll", scn.animax_items, "index", rows = 5)
        col = row.column(align=True)
        col.menu("ANIMAX_MT_items", text = '', icon = 'ADD')
        col.operator("animax.remove_item", icon='REMOVE', text = '')
        col.separator()
        col.menu("ANIMAX_MT_items_options", text = '', icon = 'DOWNARROW_HLT')
        col.separator()
        col.operator("animax.move_item", text = '', icon = 'TRIA_UP').direction = 'UP'
        col.operator("animax.move_item", text = '', icon = 'TRIA_DOWN').direction = 'DOWN'

# Item Properties Panel
class ANIMAX_PT_items_props(TemplateItems, Panel):
    bl_label = "Properties"
    bl_parent_id = "ANIMAX_PT_items"
        
    def draw(self, context):
        layout = self.layout
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return
        coll, index = items_coll(context)
        item = coll[index]
#        container = item.get('geo_node_container')
#        layout.prop(item, 'geo_node_container')
            
        if not is_container_valid(item):
            layout.alert = True
            layout.label(text = 'Container Object Missing', icon = 'ERROR')
            return
        ng = item_ng(item)
        items_dict[item.type][2](layout, item, ng)
        
# Item FX Exposed Properties
class ANIMAX_PT_items_fx_props(TemplateItems, Panel):
    bl_label = "Exposed Parameters"
    bl_parent_id = "ANIMAX_PT_items"
    bl_options = {"DEFAULT_CLOSED"}
    
    def draw_header(self, context):
        self.layout.operator("animax.exposed_prop_options", text="", icon = 'OPTIONS')
    
    def draw(self, context):
        layout = self.layout
        #layout.use_property_split = True
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return
        coll, index = items_coll(context)
        item = coll[index]
        if not len(item.exposed_fx):
            box = layout.box()
            box.label(text = "Right click on any effect's parameter", icon = 'INFO')
            box.label(text = "Then choose (Expose these parameter)", icon = 'BLANK1')
            box.label(text = "To add it here", icon = 'BLANK1')
            return
        preferences = prefs(context)
        col = layout.column(align = True)
        for idx, i in enumerate(item.exposed_fx):
            fx = item.fx.get(i.fx)
            if fx is None:
                continue
            box = col.box()
            row = box.row(align = True)
            if type(getattr(fx, i.id)) == bool:
                row.prop(fx, i.id, text = i.ui_name, icon = 'BLANK1')
            elif type(getattr(fx, i.id)) in (Vector, Euler):
                col2 = row.column()
                col2.prop(fx, i.id, text = i.ui_name)
            else:
                row.prop(fx, i.id, text = i.ui_name)
            if preferences.info_button:
                row.operator("animax.exposed_prop_info", text = "", icon = 'INFO').index = idx
            if preferences.move_button:
                move_up = row.operator("animax.move_exposed_prop", text = "", icon = 'TRIA_UP')
                move_up.direction = 'UP'
                move_up.index = idx
                move_down = row.operator("animax.move_exposed_prop", text = "", icon = 'TRIA_DOWN')
                move_down.direction = 'DOWN'
                move_down.index = idx
            if preferences.remove_button:
                row.operator("animax.remove_exposed_prop", text = "", icon = 'X').index = idx
        
# Effects Panel
class ANIMAX_PT_fx(TemplateItems, Panel):
    bl_label = "Effects"
    bl_parent_id = "ANIMAX_PT_main"
    
    @classmethod
    def poll(self, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        if not hasattr(scn, 'animax_exposed_fx'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        item = coll[index]
        if not is_container_valid(item):
            return False
        return scn.animax_tabs == 'ANIMATE'
        
    def draw(self, context):
        scn = context.scene
        layout = self.layout
        coll, index = items_coll(context)
        item = coll[index]
        
        fx = item.fx
        fx_index = item.fx_index
        
        a_icon = get_custom_icon_id('a_preset.png')
        c_icon = get_custom_icon_id('c_preset.png')
        
        col = layout.column(align = True)
        row = col.row(align = True)
        row.prop(scn, 'animax_a_presets', text = '', icon_value = a_icon)
        row.prop(scn, 'animax_c_presets', text = '', icon_value = c_icon)
        row.prop(scn, 'animax_presets_filter', text = '', icon = 'VIEWZOOM')
        
        row = col.row(align = True)
        row.prop(scn, 'animax_fx_presets', text = '')
        row.operator("animax.load_fx_preset", icon='FILE_PARENT', text = '')
        row.operator("animax.save_fx_preset", icon='ADD', text = '')
        row.operator("animax.delete_fx_preset", icon='REMOVE', text = '')
        
        row = layout.row()
        row.template_list("SCENE_UL_animax_fx", "fx", item, "fx", item, "fx_index", rows = 5)
        col = row.column(align=True)
        col.menu("ANIMAX_MT_effects", text = '', icon = 'ADD')
        col.operator("animax.remove_fx", icon='REMOVE', text = '')
        col.separator()
        col.menu("ANIMAX_MT_fx_options", text = '', icon = 'DOWNARROW_HLT')
        col.separator()
        col.operator("animax.move_fx", text = '', icon = 'TRIA_UP').direction = 'UP'
        col.operator("animax.move_fx", text = '', icon = 'TRIA_DOWN').direction = 'DOWN'
            
# Effects Offset Panel
class ANIMAX_PT_offset_n_timing(TemplateFX, Panel):
    bl_label = "Offset & Timing"
    bl_parent_id = "ANIMAX_PT_fx"
    
    def draw(self, context):
        scn = context.scene
        layout = self.layout
        coll, index = items_coll(context)
        item = coll[index]
        
        fx = item.fx
        fx_index = item.fx_index
        
        if not len(fx):
            return
        invalid_sync_fx = False
        effect = fx[fx_index]
        row = layout.row()        
        
        if effect.sync and fx_index > 0 and effect.sync_with:
            sync_index = get_item_idx(fx, effect.sync_with)
            if sync_index is None:
                invalid_sync_fx = True
            elif sync_index >= fx_index:
                invalid_sync_fx = True
        row.alert = invalid_sync_fx
        if not effect.sync or fx_index == 0:
            row.prop(effect, "offset", expand = True)
        else:
            row.prop_search(effect, "sync_with", item, "fx", text = '')
        if fx_index > 0:
            row.prop(effect, "sync", icon = 'TEMP', text = '')
        if invalid_sync_fx:
            col = layout.column()
            col.alert = invalid_sync_fx
            col.label(text = 'Only the previous effects are valid', icon = 'ERROR')
                
            
                    
# Effects Offset Panel
class ANIMAX_PT_offset(TemplateItems, Panel):
    bl_label = "Offset"
    bl_parent_id = "ANIMAX_PT_offset_n_timing"
    
    @classmethod
    def poll(self, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        fx = coll[index].fx
        fx_index = coll[index].fx_index
        if not len(fx):
            return False
        effect = fx[fx_index]
        if effect.sync and fx_index > 0:
            return False
        return scn.animax_tabs == 'ANIMATE'
        
    def draw(self, context):
        scn = context.scene
        layout = self.layout
        coll, index = items_coll(context)
        item = coll[index]
        
        fx = item.fx
        fx_index = item.fx_index
        effect = fx[fx_index]
        
        if effect.offset == 'POSITION':
            col = layout.column(align = True)
            row = col.row()
            row.prop(effect, "offset_direction", expand = True)
            col.prop(effect, "invert_direction", icon = 'ARROW_LEFTRIGHT')
            layout.prop(effect, 'offset_falloff')
            layout.prop(effect, 'random_offset')
            layout.prop(effect, 'random_offset_seed')
        elif effect.offset == 'EFFECTOR':
            col = layout.column(align = True)
            row = col.row()
            row.prop(effect, "effector_type", expand = True)
            col.prop(effect, "invert_direction", icon = 'ARROW_LEFTRIGHT')
            layout.prop(effect, 'effector_object', text = '')
            layout.prop(effect, 'use_effector_scale')
            if not effect.use_effector_scale:
                layout.prop(effect, 'effector_distance')
            layout.prop(effect, 'effector_falloff')
        elif effect.offset == 'TEXTURE':
            col = layout.column(align = True)
            row = col.row()
            row.prop(effect, "texture_type", expand = True)
            col.prop(effect, "invert_direction", icon = 'ARROW_LEFTRIGHT')
            if effect.texture_type == 'IMAGE':
                row = layout.row(align = True)
                row.prop(effect, 'image_texture', text = '')
                row.operator("animax.load_image", text = '', icon = 'FILEBROWSER')
            layout.prop(effect, 'offset_falloff', text = 'Falloff')
            if effect.texture_type in ['NOISE', 'LINEAR', 'RADIAL']:
                col = layout.column()
                col.prop(effect, 'texture_rotation')
            if effect.texture_type == 'NOISE':
                layout.prop(effect, 'noise_texture_scale')
                layout.prop(effect, 'noise_texture_detail')
                layout.prop(effect, 'noise_texture_roughness')
                layout.prop(effect, 'noise_texture_distortion')
                layout.prop(effect, 'noise_texture_frequency')
                layout.prop(effect, 'random_offset_seed')
            elif effect.texture_type == 'IMAGE':
                col = layout.column()
                col.prop(effect, 'texture_location')
                col.prop(effect, 'texture_rotation')
                col.prop(effect, 'texture_scale')
                
# Effects Timing Panel
class ANIMAX_PT_timing(TemplateItems, Panel):
    bl_label = "Timing"
    bl_parent_id = "ANIMAX_PT_offset_n_timing"
    
    @classmethod
    def poll(self, context):
        scn = context.scene
        if not hasattr(scn, 'animax_items'):
            return False
        coll, index = items_coll(context)
        if not len(coll):
            return False
        fx = coll[index].fx
        fx_index = coll[index].fx_index
        if not len(fx):
            return False
        effect = fx[fx_index]
        if effect.sync and fx_index > 0:
            return False
        if effect.offset not in ['POSITION', 'TEXTURE']:
            return False
        return scn.animax_tabs == 'ANIMATE'
        
    def draw(self, context):
        scn = context.scene
        layout = self.layout
        coll, index = items_coll(context)
        item = coll[index]
        
        fx = item.fx
        fx_index = item.fx_index
        effect = fx[fx_index]
        
        col = layout.column(align = True)
        col.prop(effect, 'control_time', icon = 'PREVIEW_RANGE')
        if effect.control_time:
            col.prop(effect, 'time', text = 'Control Time', slider = True)
        else:
            layout.prop(effect, 'start_frame')
            layout.prop(effect, 'duration')
        layout.prop(effect, 'invert_time', icon = 'ARROW_LEFTRIGHT')
        if effect.offset == 'POSITION':
            layout.prop(effect, 'repeat_time', icon = 'RECOVER_LAST')
            layout.prop(effect, 'ping_pong', icon = 'INVERSESQUARECURVE')
        if not effect.control_time:
            end_frame = effect.start_frame + effect.duration
            layout.label(text = 'End Frame: ' + str(end_frame), icon = 'INFO')
            
# Effects Interpolation Panel
class ANIMAX_PT_interpolation(TemplateFX, Panel):
    bl_label = "Interpolation"
    bl_parent_id = "ANIMAX_PT_offset_n_timing"
    
    def draw(self, context):
        scn = context.scene
        layout = self.layout
        coll, index = items_coll(context)
        item = coll[index]
        
        fx = item.fx
        fx_index = item.fx_index
        effect = fx[fx_index]
        
        row = layout.row()
        row.prop(effect, "easing_type", expand = True)
        if effect.easing_type == 'FUNCTION':
            layout.prop(effect, 'easing', text = '')
            if not effect.easing in ['INVERT', 'LINEAR']:
                if not 'BOUNCE' in effect.easing:
                    layout.prop(effect, 'easing_smooth_step')
            if 'ELASTIC' in effect.easing:
                layout.prop(effect, 'easing_frequency')
                layout.prop(effect, 'easing_amplitude')
        elif effect.easing_type == 'PING_PONG':
            layout.prop(effect, 'easing_smooth_step')
        else:
            curve_node = get_curve_node(item, effect)
            if curve_node is not None:
                layout.template_curve_mapping(curve_node, "mapping")
                layout.prop(effect, 'easing_curve_factor', slider = True)
                
# Effects Properties Panel
class ANIMAX_PT_fx_props(TemplateFX, Panel):
    bl_label = "Properties"
    bl_parent_id = "ANIMAX_PT_fx"
    
    def draw(self, context):
        pass
        
# Effects Location Panel
class ANIMAX_PT_fx_location(TemplateFX, Panel):
    bl_label = "Location"
    bl_parent_id = "ANIMAX_PT_fx_props"
    
    def draw_header(self,context):
        coll, index = items_coll(context)
        item = coll[index]
        fx = item.fx
        fx_index = item.fx_index
        effect = fx[fx_index]
        
        additive = get_custom_icon_id("add.png")
        replace = get_custom_icon_id("replace.png")
        icon = additive if is_fx_loc_additive(effect) else replace
        if icon is not None:
            self.layout.label(text = '', icon_value = icon)
        
    def draw(self, context):
        scn = context.scene
        layout = self.layout
        coll, index = items_coll(context)
        item = coll[index]
        
        fx = item.fx
        fx_index = item.fx_index
        
        if not len(fx):
            return
        effect = fx[fx_index]
        effects_dict[effect.type][2](layout, item, effect, 'LOC')
        
# Effects Rotation Panel
class ANIMAX_PT_fx_rotation(TemplateFX, Panel):
    bl_label = "Rotation"
    bl_parent_id = "ANIMAX_PT_fx_props"
    
    def draw_header(self,context):
        coll, index = items_coll(context)
        item = coll[index]
        fx = item.fx
        fx_index = item.fx_index
        effect = fx[fx_index]
        
        additive = get_custom_icon_id("add.png")
        replace = get_custom_icon_id("replace.png")
        icon = additive if is_fx_rot_additive(effect) else replace
        if icon is not None:
            self.layout.label(text = '', icon_value = icon)
        
    def draw(self, context):
        scn = context.scene
        layout = self.layout
        coll, index = items_coll(context)
        item = coll[index]
        
        fx = item.fx
        fx_index = item.fx_index
        
        if not len(fx):
            return
        effect = fx[fx_index]
        effects_dict[effect.type][2](layout, item, effect, 'ROT')
        
# Effects Scale Panel
class ANIMAX_PT_fx_scale(TemplateFX, Panel):
    bl_label = "Scale"
    bl_parent_id = "ANIMAX_PT_fx_props"
    
    def draw_header(self,context):
        coll, index = items_coll(context)
        item = coll[index]
        fx = item.fx
        fx_index = item.fx_index
        effect = fx[fx_index]
        
        additive = get_custom_icon_id("add.png")
        replace = get_custom_icon_id("replace.png")
        icon = additive if is_fx_scale_additive(effect) else replace
        if icon is not None:
            self.layout.label(text = '', icon_value = icon)
        
    def draw(self, context):
        scn = context.scene
        layout = self.layout
        coll, index = items_coll(context)
        item = coll[index]
        
        fx = item.fx
        fx_index = item.fx_index
        
        if not len(fx):
            return
        effect = fx[fx_index]
        effects_dict[effect.type][2](layout, item, effect, 'SCALE')
        
# Effects Extra options Panel
class ANIMAX_PT_fx_extra(TemplateFX, Panel):
    bl_label = "Extra Parameters"
    bl_parent_id = "ANIMAX_PT_fx_props"
        
    def draw(self, context):
        scn = context.scene
        layout = self.layout
        coll, index = items_coll(context)
        item = coll[index]
        
        fx = item.fx
        fx_index = item.fx_index
        
        if not len(fx):
            return
        effect = fx[fx_index]
        effects_dict[effect.type][2](layout, item, effect, 'EXTRA')

        
############################################################################################
################################ Register/Unregister #######################################
############################################################################################

classes = (
    ANIMAX_Preferences,
    ANIMAX_PT_main,
    ANIMAX_PT_items,
    ANIMAX_PT_items_props,
    ANIMAX_PT_items_fx_props,
    ANIMAX_PT_fx,
    ANIMAX_PT_offset_n_timing,
    ANIMAX_PT_offset,
    ANIMAX_PT_timing,
    ANIMAX_PT_interpolation,
    ANIMAX_PT_fx_props,
    ANIMAX_PT_fx_location,
    ANIMAX_PT_fx_rotation,
    ANIMAX_PT_fx_scale,
    ANIMAX_PT_fx_extra,
    SCENE_UL_animax_items,
    SCENE_UL_animax_fx,
    ANIMAX_OT_load_props,
    ANIMAX_OT_add_item,
    ANIMAX_OT_remove_item,
    ANIMAX_OT_items_visibility,
    ANIMAX_OT_remove_fx,
    ANIMAX_OT_dplicate_fx,
    ANIMAX_OT_dplicate_item,
    ANIMAX_OT_add_fx,
    ANIMAX_OT_move_item,
    ANIMAX_OT_move_fx,
    ANIMAX_OT_move_exposed_prop,
    ANIMAX_OT_enable_items,
    ANIMAX_OT_enable_fx,
    ANIMAX_OT_remove_all_items,
    ANIMAX_OT_remove_all_fx,
    ANIMAX_OT_load_font,
    ANIMAX_OT_load_image,
    ANIMAX_OT_copy_name,
    ANIMAX_OT_save_fx_preset,
    ANIMAX_OT_load_fx_preset,
    ANIMAX_OT_delete_fx_preset,
    ANIMAX_OT_open_manual,
    ANIMAX_OT_yt_playlist,
    ANIMAX_OT_save_clipboard,
    ANIMAX_OT_load_clipboard,
    ANIMAX_OT_bake_to_shapekeys,
    ANIMAX_OT_bake_to_keyframes,
    ANIMAX_OT_expose_fx_prop,
    ANIMAX_OT_remove_exposed_prop,
    ANIMAX_OT_exposed_prop_info,
    ANIMAX_OT_exposed_prop_options,
    ANIMAX_OT_check_mesh,
    ANIMAX_OT_load_generators_ng,
    AnimaxItems,
    AnimaxEffects,
    Animax_Items_Options,
    Animax_FX_Options,
)

def register():
    from bpy.utils import register_class, previews
    for cls in classes:
        register_class(cls)
    bpy.types.Scene.animax_tabs = EnumProperty(
        items = (
            ('ANIMATE', 'Animate', 'Create procedural animations', 'ONIONSKIN_ON', 0),
            ('FRACTURE', 'Fracture', 'Fracture objects to multiple pieces', 'MOD_PHYSICS', 1)
        ),
        description = 'What you want to do?'
    )
    bpy.types.Scene.animax_fx_presets = EnumProperty(
        name = 'Presets',
        items = animation_presets_enum,
        description = 'Available Presets'
    )
    bpy.types.Scene.animax_a_presets = BoolProperty(
        name = 'Built-in Presets',
        default = True,
        update = update_presets,
        description = 'Display the presets that are shipped with the add-on'
    )
    bpy.types.Scene.animax_c_presets = BoolProperty(
        name = 'Custom Presets',
        default = True,
        update = update_presets,
        description = 'Display the custom presets'
    )
    bpy.types.Scene.animax_presets_filter = StringProperty(
        name = 'Presets Filter',
        update = update_presets,
        description = 'Live search filtering string',
        options = {'TEXTEDIT_UPDATE'}
    )
    bpy.types.Scene.animax_fracture_tabs = EnumProperty(
        items = (
            ('SPLIT', 'Split', 'Split the shell of the object'),
            ('SLICE', 'Slice', 'Slice the object'),
            ('VORONOI', 'Voronoi', 'Voronoi fracturing')
        ),
        description = 'Fracturing Options'
    )
    bpy.types.Scene.animax_update_tag = BoolProperty(update = update_fx_node_tree)
    
    # Previews for the icons
    pcoll = previews.new()
    for i in os.listdir(icons_dir):
        if not i.lower().endswith('.png'):
            continue
        pcoll.load(i, os.path.join(icons_dir, i), 'IMAGE')
    preview_collections["icons"] = pcoll
    
    # Add to right click menu
    bpy.types.UI_MT_button_context_menu.append(draw_menu)
    
    load_post.append(animax_load_handler)
    
    animax_shell_splitting.register()
    animax_slicing.register()
    animax_voronoi_fracture.register()
    
def unregister():
    from bpy.utils import unregister_class, previews
    for cls in reversed(classes):
        unregister_class(cls)
        
    del bpy.types.Scene.animax_tabs
    del bpy.types.Scene.animax_fx_presets
    del bpy.types.Scene.animax_a_presets
    del bpy.types.Scene.animax_c_presets
    del bpy.types.Scene.animax_presets_filter
    del bpy.types.Scene.animax_fracture_tabs
    del bpy.types.Scene.animax_update_tag
    
    for pcoll in preview_collections.values():
        previews.remove(pcoll)
    preview_collections.clear()
    
    bpy.types.UI_MT_button_context_menu.remove(draw_menu)
    
    load_post.remove(animax_load_handler)
    
    animax_shell_splitting.unregister()
    animax_slicing.unregister()
    animax_voronoi_fracture.unregister()

if __name__ == "__main__":
    register()